# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
from numba_cuda_mlir.errors import ensure_verifies
from numba_cuda_mlir.descriptor import MLIRDispatcher
from numba_cuda_mlir.descriptor import MLIRDispatcherType
from numba_cuda_mlir.errors import InternalCompilerError
from functools import singledispatch
from dataclasses import dataclass
from abc import abstractmethod
from collections.abc import Sequence
import functools
import numpy as np
from numba_cuda_mlir.numba_cuda import types, typing
from numba_cuda_mlir.annotations import AnyCallable, PS
from numba_cuda_mlir.lowering_utilities import type_conversions
from numba_cuda_mlir.lowering_utilities.type_conversions import (
    np_dtype_to_mlir_type as mlir_type_from_numpy_dtype,
    to_mlir_type,
)
from numba_cuda_mlir.type_defs.float_types import BFloat16Type
from numba_cuda_mlir._mlir import ir
from numba_cuda_mlir._mlir.dialects import (
    arith,
    memref,
    func,
    gpu,
    llvm,
    cf,
    builtin,
    complex as complex_dialect,
    vector as vector_dialect,
)
from numba_cuda_mlir.mlir.dialect_exts import scf
from numba_cuda_mlir._mlir.extras.meta import region_op
from numba_cuda_mlir.logging import trace
from numba_cuda_mlir._mlir.extras import types as T
import operator

# std::numeric_limits<int32_t>::min() — marks a dynamic index in llvm.getelementptr
GEP_DYNAMIC_INDEX = -2147483648


def memref_llvm_address_space(memref_type: ir.MemRefType) -> int:
    """Return the NVVM LLVM address space for a CUDA memref type.

    Untagged CUDA kernel argument memrefs represent global memory. Tagged memrefs
    are used for explicit CUDA spaces such as workgroup/shared and private/local.
    """
    memory_space = memref_type.memory_space
    if memory_space is None:
        return 1
    if isinstance(memory_space, ir.IntegerAttr):
        return int(memory_space.value)
    memory_space_str = str(memory_space)
    if "workgroup" in memory_space_str:
        return 3
    if "private" in memory_space_str:
        return 5
    return 1


def _memref_llvm_pointer_type(memref_type: ir.MemRefType) -> llvm.PointerType:
    return llvm.PointerType.get(memref_llvm_address_space(memref_type))


def memref_data_pointer_as_index(array: ir.Value, element_type: ir.Type | None = None) -> ir.Value:
    metadata = memref.extract_strided_metadata(array)
    base_ptr_idx = memref.extract_aligned_pointer_as_index(metadata[0])
    offset = index_of(metadata[1])
    if element_type is None:
        element_type = ir.MemRefType(array.type).element_type
    elem_bytes = get_type_size_bytes(element_type)
    byte_offset = arith.muli(offset, arith.constant(T.index(), elem_bytes))
    return arith.addi(base_ptr_idx, byte_offset)


def _memref_index_offset(array: ir.Value, indices: list[ir.Value]) -> ir.Value:
    """Return the element offset for indices into a potentially-strided memref."""
    metadata = memref.extract_strided_metadata(array)
    rank = array.type.rank
    ndim = len(indices)
    if ndim == 1 and rank != 1:
        return convert(indices[0], T.i64())
    if ndim != rank:
        raise ValueError(
            f"Expected either a scalar linear index or {rank} indices for {array.type}, got {ndim}"
        )
    linear_idx = arith.constant(T.i64(), 0)
    for d, index in enumerate(indices):
        idx_val = convert(index, T.i64())
        stride = convert(metadata[2 + rank + d], T.i64())
        linear_idx = linear_idx + idx_val * stride
    return linear_idx


def memref_to_llvm_ptr(array: ir.Value, indices: list[ir.Value], element_type: ir.Type) -> ir.Value:
    """Convert memref + indices to LLVM pointer.

    Extracts base pointer from a potentially-strided memref and computes the
    element pointer using the memref metadata offset and strides.

    Args:
        array: Memref value (potentially strided)
        indices: List of index values
        element_type: Element type for getelementptr

    Returns:
        LLVM pointer (!llvm.ptr) to the indexed element
    """
    # Extract base pointer from memref and convert to an address-space-preserving
    # LLVM pointer.
    ptr_type = _memref_llvm_pointer_type(ir.MemRefType(array.type))
    base_ptr_idx = memref_data_pointer_as_index(array)
    base_ptr = llvm.inttoptr(res=ptr_type, arg=convert(base_ptr_idx, T.i64()))

    linear_idx = _memref_index_offset(array, indices)
    return llvm.getelementptr(
        ptr_type,
        base_ptr,
        [linear_idx],
        [GEP_DYNAMIC_INDEX],
        to_llvm_storable_type(element_type),
        None,
    )


def to_llvm_storable_type(element_type: ir.Type) -> ir.Type:
    if is_complex_type(element_type):
        return get_llvm_struct_for_complex(element_type)
    return element_type


def llvm_ptr_load(element_type: ir.Type, ptr: ir.Value) -> ir.Value:
    if is_complex_type(element_type):
        loaded = llvm.load(get_llvm_struct_for_complex(element_type), ptr)
        return llvm_struct_to_complex(loaded, element_type)
    return llvm.load(element_type, ptr)


def llvm_ptr_store(value: ir.Value, ptr: ir.Value) -> None:
    if is_complex_type(value.type):
        value = complex_to_llvm_struct(value)
    llvm.store(value, ptr)


def memref_to_tensor(memref):
    """
    Utility function to convert from a memref to tensor
    """
    from numba_cuda_mlir._mlir.dialects import bufferization

    match memref.type:
        case ir.MemRefType():
            tensor_type = ir.RankedTensorType.get(
                shape=memref.type.shape, element_type=memref.type.element_type
            )
            tensor = bufferization.to_tensor(result=tensor_type, buffer=memref, restrict=True)
            return tensor
        case ir.RankedTensorType():
            return memref
        case _:
            raise NotImplementedError(f"Not implemented for type {memref.type}")


def tensor_to_memref(tensor):
    """
    Utility function to convert from a tensor to memref
    """
    from numba_cuda_mlir._mlir.dialects import bufferization

    match tensor.type:
        case ir.MemRefType():
            return tensor
        case ir.RankedTensorType():
            # Create memref with strided layout
            rank = len(tensor.type.shape)
            dyn_stride = ir.MemRefType.get_dynamic_stride_or_offset()
            layout = ir.StridedLayoutAttr.get(offset=dyn_stride, strides=[dyn_stride] * rank)
            memref_type = ir.MemRefType.get(
                shape=tensor.type.shape,
                element_type=tensor.type.element_type,
                layout=layout,
            )
            memref = bufferization.to_buffer(buffer=memref_type, tensor=tensor)
            return memref
        case _:
            raise NotImplementedError(f"Not implemented for type {tensor.type}")


def get_type_width(ty: ir.Type) -> int:
    match ty:
        case ir.IntegerType():
            return ty.width
        case ir.IndexType():
            return 64
        case ir.FloatType():
            return ty.width
        case ir.ComplexType():
            return 2 * ir.ComplexType(ty).element_type.width
        case ir.VectorType():
            vt = ir.VectorType(ty)
            count = 1
            for d in vt.shape:
                count *= d
            return count * get_type_width(vt.element_type)
        case _:
            raise NotImplementedError(f"Not implemented for type {ty}")


def is_complex_type(mlir_type):
    return isinstance(mlir_type, ir.ComplexType)


# Complex values lower as MLIR complex scalars in SSA, but pointer and record
# storage paths address them as literal LLVM {real, imag} structs. Keep the
# conversions shared so those independent lowering paths use the same layout.
def get_llvm_struct_for_complex(complex_type):
    elem_type = complex_type.element_type
    return llvm.StructType.get_literal([elem_type, elem_type])


def complex_to_llvm_struct(value):
    complex_type = value.type
    if not is_complex_type(complex_type):
        raise TypeError(f"Expected MLIR complex type, got {complex_type}")
    struct_type = get_llvm_struct_for_complex(complex_type)
    real = complex_dialect.re(value)
    imag = complex_dialect.im(value)
    undef = llvm.UndefOp(struct_type)
    with_real = llvm.insertvalue(
        container=undef,
        value=real,
        position=ir.DenseI64ArrayAttr.get([0]),
    )
    return llvm.insertvalue(
        container=with_real,
        value=imag,
        position=ir.DenseI64ArrayAttr.get([1]),
    )


def llvm_struct_to_complex(value, complex_type):
    elem_type = complex_type.element_type
    real = llvm.extractvalue(
        res=elem_type,
        container=value,
        position=ir.DenseI64ArrayAttr.get([0]),
    )
    imag = llvm.extractvalue(
        res=elem_type,
        container=value,
        position=ir.DenseI64ArrayAttr.get([1]),
    )
    return complex_dialect.create_(complex=complex_type, real=real, imaginary=imag)


def get_type_size_bytes(ty: ir.Type) -> int:
    if isinstance(ty, ir.BF16Type):
        return 2
    # CUDA memrefs handled here are byte-addressable. Numpy bool storage is one
    # byte even though the MLIR element type is i1.
    return max(1, (get_type_width(ty) + 7) // 8)


def _lookup_datamodel_type(numba_type: types.Type, context: str) -> ir.Type:
    from numba_cuda_mlir.models import mlir_data_manager

    model = mlir_data_manager.lookup(numba_type)
    match context:
        case "value":
            return model.get_value_type()
        case "storage" | "data":
            return model.get_data_type()
        case "argument":
            return model.get_argument_type()
        case "return":
            return model.get_return_type()
        case _:
            raise InternalCompilerError(f"Unknown datamodel context: {context}")


def get_value_type(numba_type: types.Type) -> ir.Type:
    return _lookup_datamodel_type(numba_type, "value")


def get_storage_type(numba_type: types.Type) -> ir.Type:
    return _lookup_datamodel_type(numba_type, "storage")


def get_argument_type(numba_type: types.Type) -> ir.Type:
    return _lookup_datamodel_type(numba_type, "argument")


def get_return_type(numba_type: types.Type) -> ir.Type:
    return _lookup_datamodel_type(numba_type, "return")


def _numba_type_bitwidth(numba_type: types.Type, value_type: ir.Type | None = None) -> int:
    bitwidth = getattr(numba_type, "bitwidth", None)
    if bitwidth is not None:
        return int(bitwidth)
    if value_type is not None:
        return get_type_width(value_type)
    return get_type_width(get_value_type(numba_type))


def storage_bitwidth(numba_type: types.Type) -> int:
    storage_type = get_storage_type(numba_type)
    return get_type_width(storage_type)


def storage_itemsize_bytes(numba_type: types.Type) -> int:
    from numba_cuda_mlir.types import NestedArray, Record

    if isinstance(numba_type, Record):
        return numba_type.size
    if isinstance(numba_type, NestedArray):
        return storage_itemsize_bytes(numba_type.dtype)

    dtype = getattr(numba_type, "dtype", numba_type)
    if isinstance(dtype, Record):
        return dtype.size
    if isinstance(dtype, NestedArray):
        return storage_itemsize_bytes(dtype.dtype)
    if isinstance(dtype, np.dtype):
        return dtype.itemsize
    if isinstance(dtype, types.UnicodeCharSeq):
        return getattr(dtype, "count", 1) * np.dtype("U1").itemsize
    if isinstance(dtype, types.CharSeq):
        return getattr(dtype, "count", 1)
    bitwidth = storage_bitwidth(dtype)
    return (bitwidth + 7) // 8


def _is_bool_numba_type(numba_type: types.Type) -> bool:
    return isinstance(numba_type, (types.Boolean, types.BooleanLiteral))


def _is_float_storage_numba_type(numba_type: types.Type) -> bool:
    from numba_cuda_mlir.type_defs import float_types
    from numba_cuda_mlir.numba_cuda.types.ext_types import Bfloat16

    return isinstance(numba_type, (types.Float, float_types.SpecialFloatType, Bfloat16))


def _integer_storage_type_for_value(numba_type: types.Type, value_type: ir.Type) -> ir.IntegerType:
    return ir.IntegerType.get_signless(_numba_type_bitwidth(numba_type, value_type))


def value_to_storage(numba_type: types.Type, value: ir.Value) -> ir.Value:
    """Convert a source-level value to its memory/ABI storage representation."""
    storage_type = get_storage_type(numba_type)
    value_type = get_value_type(numba_type)
    if getattr(value, "type", None) == storage_type and value_type == storage_type:
        return value
    if getattr(value, "type", None) != value_type:
        value = convert(value, value_type)
    if value_type == storage_type:
        return value

    if _is_bool_numba_type(numba_type):
        value = convert(value, T.bool())
        return arith.extui(out=storage_type, in_=value)

    if _is_float_storage_numba_type(numba_type) and isinstance(storage_type, ir.IntegerType):
        bits_type = _integer_storage_type_for_value(numba_type, value_type)
        bits = arith.bitcast(out=bits_type, in_=value)
        if bits_type.width == storage_type.width:
            return bits if bits_type == storage_type else arith.bitcast(out=storage_type, in_=bits)
        if bits_type.width < storage_type.width:
            return arith.extui(out=storage_type, in_=bits)
        return arith.trunci(out=storage_type, in_=bits)

    return convert(value, storage_type)


def storage_to_value(numba_type: types.Type, value: ir.Value) -> ir.Value:
    """Convert a memory/ABI storage payload to its source-level value representation."""
    storage_type = get_storage_type(numba_type)
    value_type = get_value_type(numba_type)
    if getattr(value, "type", None) == value_type and value_type == storage_type:
        return value
    if getattr(value, "type", None) != storage_type:
        value = convert(value, storage_type)
    if value_type == storage_type:
        return value

    if _is_bool_numba_type(numba_type):
        zero = arith.constant(storage_type, 0)
        return arith.cmpi(arith.CmpIPredicate.ne, value, zero)

    if _is_float_storage_numba_type(numba_type) and isinstance(storage_type, ir.IntegerType):
        bits_type = _integer_storage_type_for_value(numba_type, value_type)
        bits = value
        if storage_type.width > bits_type.width:
            bits = arith.trunci(out=bits_type, in_=value)
        elif storage_type.width < bits_type.width:
            bits = arith.extui(out=bits_type, in_=value)
        elif value.type != bits_type:
            bits = arith.bitcast(out=bits_type, in_=value)
        return arith.bitcast(out=value_type, in_=bits)

    return convert(value, value_type)


def array_element_value_load(
    array_type: types.Array,
    array: ir.Value,
    indices: Sequence[ir.Value],
    *,
    dynamic_shared_memory: bool = False,
):
    if dynamic_shared_memory:
        storage_type = get_storage_type(array_type.dtype)
        ptr = memref_to_llvm_ptr(array, list(indices), storage_type)
        stored = llvm_ptr_load(storage_type, ptr)
    else:
        stored = memref.load(array, list(indices))
    return storage_to_value(array_type.dtype, stored)


def array_element_value_store(
    array_type: types.Array,
    array: ir.Value,
    indices: Sequence[ir.Value],
    value: ir.Value,
    *,
    dynamic_shared_memory: bool = False,
):
    stored = value_to_storage(array_type.dtype, value)
    if dynamic_shared_memory:
        ptr = memref_to_llvm_ptr(array, list(indices), stored.type)
        llvm_ptr_store(stored, ptr)
    else:
        memref.store(value=stored, memref=array, indices=list(indices))


def memref_to_value_tensor(array_type: types.Array, array: ir.Value) -> ir.Value:
    from numba_cuda_mlir._mlir.dialects import linalg, tensor

    tensor_value = memref_to_tensor(array)
    value_element_type = get_value_type(array_type.dtype)
    if tensor_value.type.element_type == value_element_type:
        return tensor_value

    result_type = T.tensor(*tensor_value.type.shape, value_element_type)
    init = tensor.empty(
        sizes=dims_of_tensor_shape(tensor_value),
        element_type=value_element_type,
    )
    result = linalg.MapOp(
        result=[result_type],
        inputs=[tensor_value],
        init=init,
    )
    block = result.mapper.blocks.append(tensor_value.type.element_type, value_element_type)
    with ir.InsertionPoint(block):
        linalg.yield_([storage_to_value(array_type.dtype, block.arguments[0])])
    return result.result[0]


def value_tensor_to_storage_memref(array_type: types.Array, tensor_value: ir.Value) -> ir.Value:
    from numba_cuda_mlir._mlir.dialects import linalg, tensor

    storage_element_type = get_storage_type(array_type.dtype)
    if tensor_value.type.element_type == storage_element_type:
        return tensor_to_memref(tensor_value)

    result_type = T.tensor(*tensor_value.type.shape, storage_element_type)
    init = tensor.empty(
        sizes=dims_of_tensor_shape(tensor_value),
        element_type=storage_element_type,
    )
    result = linalg.MapOp(
        result=[result_type],
        inputs=[tensor_value],
        init=init,
    )
    block = result.mapper.blocks.append(tensor_value.type.element_type, storage_element_type)
    with ir.InsertionPoint(block):
        linalg.yield_([value_to_storage(array_type.dtype, block.arguments[0])])
    return tensor_to_memref(result.result[0])


@singledispatch
def numpy_implicit_type_promotion(
    a: ir.Type | types.Type, b: ir.Type | types.Type
) -> ir.Type | types.Type:
    raise NotImplementedError(f"{a=} {b=}")


@numpy_implicit_type_promotion.register(types.Type)
def _(a: types.Type, b: types.Type) -> types.Type:
    float_types = {
        16: types.float16,
        32: types.float32,
        64: types.float64,
    }
    integer_types = {
        8: types.int8,
        16: types.int16,
        32: types.int32,
        64: types.int64,
    }
    from numba_cuda_mlir.type_defs.float_types import SpecialFloatType

    match a, b:
        case ty1, ty2 if ty1 == ty2:
            return a
        case (BFloat16Type(), _ as other) | (_ as other, BFloat16Type()):
            raise TypeError(f"Trying to convert to/from BFloat16 and {other}")
        case (SpecialFloatType() as s, _ as other) | (
            _ as other,
            SpecialFloatType() as s,
        ):
            raise TypeError(f"Implicit promotion between {s} and {other} is not supported")
        case types.Float() as f, types.Integer() as i:
            return float_types[max(f.bitwidth, i.bitwidth)]
        case types.Integer() as i, types.Float() as f:
            return float_types[max(f.bitwidth, i.bitwidth)]
        case types.Integer() as i1, types.Integer() as i2:
            return integer_types[max(i1.bitwidth, i2.bitwidth)]
        case types.Float() as f1, types.Float() as f2:
            return float_types[max(f1.bitwidth, f2.bitwidth)]
        case _:
            raise NotImplementedError(f"{a=} {b=}")


@numpy_implicit_type_promotion.register(ir.Type)
def _(a: ir.Type, b: ir.Type) -> ir.Type:
    match a, b:
        case ty1, ty2 if ty1 == ty2:
            return a
        case (
            (ir.IntegerType(), ir.FloatType())
            | (ir.FloatType(), ir.IntegerType())
            | (ir.IndexType(), ir.FloatType())
            | (ir.FloatType(), ir.IndexType())
        ):
            larger_width = max(get_type_width(a), get_type_width(b))
            ty = T.f32() if larger_width == 32 else T.f64()
            return ty
        case (
            (ir.IntegerType(), ir.IntegerType())
            | (ir.IndexType(), ir.IntegerType())
            | (ir.IntegerType(), ir.IndexType())
        ):
            larger_width = max(get_type_width(a), get_type_width(b))
            ty = ir.IntegerType.get_signless(width=larger_width)
            return ty
        case ir.FloatType(), ir.FloatType():
            larger_width = max(get_type_width(a), get_type_width(b))
            ty = T.f32() if larger_width == 32 else T.f64()
            return ty
        case (ir.ComplexType(), _) | (_, ir.ComplexType()):

            def _scalar_width(t: ir.Type) -> int:
                return t.element_type.width if isinstance(t, ir.ComplexType) else get_type_width(t)

            larger_width = max(_scalar_width(a), _scalar_width(b))
            elem = T.f32() if larger_width <= 32 else T.f64()
            return T.complex(elem)
        case _:
            raise NotImplementedError(f"Not implemented for type {a} and {b}")


def coerce_numpy_scalars_for_binary_op(a: ir.Value, b: ir.Value) -> tuple[ir.Value, ir.Value]:
    coerced = numpy_implicit_type_promotion(a.type, b.type)
    a_signed = isinstance(a.type, ir.IntegerType) and a.type.width > 1
    b_signed = isinstance(b.type, ir.IntegerType) and b.type.width > 1
    return convert(a, coerced, signed=a_signed), convert(b, coerced, signed=b_signed)


def mul(a: ir.Value, b: ir.Value) -> ir.Value:
    """Multiply two values with automatic type coercion."""
    a, b = coerce_numpy_scalars_for_binary_op(a, b)
    match a.type:
        case ir.IntegerType() | ir.IndexType():
            return arith.muli(a, b)
        case ir.FloatType():
            return arith.mulf(a, b)
        case _:
            raise NotImplementedError(f"Not implemented for type {a.type}")


def div(a: ir.Value, b: ir.Value) -> ir.Value:
    """Divide two values with automatic type coercion."""
    a, b = coerce_numpy_scalars_for_binary_op(a, b)
    match a.type:
        case ir.IntegerType() | ir.IndexType():
            return arith.divsi(a, b)  # signed division
        case ir.FloatType():
            return arith.divf(a, b)
        case _:
            raise NotImplementedError(f"Not implemented for type {a.type}")


def less_than(a: ir.Value, b: ir.Value) -> ir.Value:
    a, b = coerce_numpy_scalars_for_binary_op(a, b)
    match a.type:
        case ir.IntegerType() | ir.IndexType():
            return arith.cmpi(arith.CmpIPredicate.slt, a, b)
        case ir.FloatType():
            return arith.cmpf(arith.CmpFPredicate.OLT, a, b)
        case _:
            raise NotImplementedError(f"Not implemented for type {a.type}")


def _get_mlir_bin_op_for_operator(op):
    import functools

    match op:
        case operator.or_:
            return (
                functools.partial(arith.ori),
                None,
            )
        case operator.and_:
            return (
                functools.partial(arith.andi),
                None,
            )
        case operator.xor:
            return (
                functools.partial(arith.xori),
                None,
            )
        case operator.lt:
            return (
                functools.partial(arith.cmpi, arith.CmpIPredicate.slt),
                functools.partial(arith.cmpf, arith.CmpFPredicate.ULT),
            )
        case operator.le:
            return (
                functools.partial(arith.cmpi, arith.CmpIPredicate.sle),
                functools.partial(arith.cmpf, arith.CmpFPredicate.ULE),
            )
        case operator.gt:
            return (
                functools.partial(arith.cmpi, arith.CmpIPredicate.sgt),
                functools.partial(arith.cmpf, arith.CmpFPredicate.UGT),
            )
        case operator.ge:
            return (
                functools.partial(arith.cmpi, arith.CmpIPredicate.sge),
                functools.partial(arith.cmpf, arith.CmpFPredicate.UGE),
            )
        case operator.eq:
            return (
                functools.partial(arith.cmpi, arith.CmpIPredicate.eq),
                functools.partial(arith.cmpf, arith.CmpFPredicate.UEQ),
            )
        case operator.ne:
            return (
                functools.partial(arith.cmpi, arith.CmpIPredicate.ne),
                functools.partial(arith.cmpf, arith.CmpFPredicate.UNE),
            )
        case operator.add | operator.iadd:
            return (
                functools.partial(arith.addi),
                functools.partial(arith.addf),
            )
        case operator.sub | operator.isub:
            return (
                functools.partial(arith.subi),
                functools.partial(arith.subf),
            )
        case operator.mul | operator.imul:
            return (
                functools.partial(arith.muli),
                functools.partial(arith.mulf),
            )
        case operator.truediv | operator.floordiv | operator.itruediv | operator.ifloordiv:
            return (
                functools.partial(arith.divsi),
                functools.partial(arith.divf),
            )
        case operator.mod | operator.imod:
            return (
                functools.partial(arith.remsi),
                functools.partial(arith.remf),
            )
        case _:
            raise NotImplementedError(f"Not implemented for operator {op}")


def _create_utility_bin_op(op):
    iop, fop = _get_mlir_bin_op_for_operator(op)

    def bin_op(a: ir.Value, b: ir.Value, *rest) -> ir.Value:
        if rest:
            return bin_op(bin_op(a, b), *rest)
        a, b = coerce_numpy_scalars_for_binary_op(a, b)
        match a.type:
            case ir.IntegerType() | ir.IndexType() if iop is not None:
                return iop(a, b)
            case ir.FloatType() if fop is not None:
                return fop(a, b)
            case _:
                raise NotImplementedError(f"Not implemented for type {a.type}")

    return bin_op


sub = _create_utility_bin_op(operator.sub)
add = _create_utility_bin_op(operator.add)
mul = _create_utility_bin_op(operator.mul)
div = _create_utility_bin_op(operator.truediv)
less_than = _create_utility_bin_op(operator.lt)
less_than_or_equal = _create_utility_bin_op(operator.le)
greater_than = _create_utility_bin_op(operator.gt)
greater_than_or_equal = _create_utility_bin_op(operator.ge)
equal = _create_utility_bin_op(operator.eq)
not_equal = _create_utility_bin_op(operator.ne)
or_ = _create_utility_bin_op(operator.or_)
and_ = _create_utility_bin_op(operator.and_)
xor = _create_utility_bin_op(operator.xor)


def true() -> ir.Value:
    return arith.constant(T.bool(), 1)


def false() -> ir.Value:
    return arith.constant(T.bool(), 0)


_generic_rmw = region_op(
    memref.GenericAtomicRMWOp,
    terminator=lambda results: memref.AtomicYieldOp(results[0]),
)


def set_error_code_if_zero(error_ptr: ir.Value, error_code: int):
    """Set error code if currently 0 (first error wins).

    Uses LLVM cmpxchg to atomically compare-and-swap the error code.
    Only the first error is recorded (subsequent errors are ignored).
    """
    zero = llvm.ConstantOp(T.i32(), ir.IntegerAttr.get(T.i32(), 0)).result
    error_val = llvm.ConstantOp(T.i32(), ir.IntegerAttr.get(T.i32(), error_code)).result
    # Atomic CAS: if current value is 0, set to error_code
    llvm.cmpxchg(
        error_ptr,
        zero,
        error_val,
        llvm.AtomicOrdering.monotonic,
        llvm.AtomicOrdering.monotonic,
    )


def set_error_and_return(
    condition: ir.Value, error_memref: ir.Value, error_code: int, return_block: ir.Block
):
    """If condition is false, set error code and branch to return block."""
    with scf.if_ctx_manager(arith.cmpi(arith.CmpIPredicate.eq, condition, false())):
        set_error_code_if_zero(error_memref, error_code)
        cf.br([], return_block)
        scf.yield_([])


def bool_of(value: ir.Value | bool) -> ir.Value:
    match value:
        case ir.Value():
            return convert(value, T.bool())
        case _:
            return true() if value else false()


def concretize_tuple_to_tensor(tup: tuple[ir.Value]) -> ir.Value:
    from numba_cuda_mlir._mlir.dialects import tensor

    ty = tup[0].type
    if not all(ty == t.type for t in tup):
        raise NotImplementedError("All elements of the tuple must have the same type")
    tensor_type = T.tensor(len(tup), ty)
    t = tensor.splat(tensor_type, tup[0], [])
    for i, element in enumerate(tup[1:]):
        t = tensor.insert(element, t, [index_of(i)])
    return t


def convert_tuple_like(values: list[ir.Value], target_type: ir.Type) -> ir.Value:
    from numba_cuda_mlir._mlir.dialects import tensor

    match target_type:
        case tuple():
            return tuple(convert(value, ty) for value, ty in zip(values, target_type))
        case ir.MemRefType():
            tty = T.tensor(*target_type.shape, target_type.element_type)
            tens = tensor.from_elements(tty, [*values])
            return tensor_to_memref(tens)
        case ir.RankedTensorType():
            return tensor.from_elements(target_type, *values)
        case _:
            raise NotImplementedError(f"Not implemented for type {target_type}")


def _convert_integer_to_integer(
    value: ir.Value, target_type: ir.IntegerType, *, signed: bool = False
) -> ir.Value:
    """
    If possible, we perform the conversion on the types as they are given to us.
    However, if the signedness of the types do not match, we first bitcast to signless forms
    and then extend/truncate as necessary.
    """
    value_type: ir.IntegerType = value.type

    if value_type == target_type:
        return value

    def _signedness(ty: ir.IntegerType) -> str:
        if ty.is_signed:
            return "signed"
        if ty.is_unsigned:
            return "unsigned"
        return "signless"

    source_signedness = _signedness(value_type)
    use_signed_extend = signed or source_signedness == "signed"

    work_value_type = value_type
    work_target_type = target_type
    if not work_target_type.is_signless:
        work_target_type = ir.IntegerType.get_signless(work_target_type.width)

    if work_value_type.width > work_target_type.width:
        # Special case: when converting to i1 (boolean), use comparison against zero
        # instead of truncation, which only keeps the LSB and fails for values like 2
        if work_target_type.width == 1:
            trace("converting to i1 (boolean) via comparison against zero")
            zero = arith.constant(work_value_type, value=0)
            value = arith.cmpi(arith.CmpIPredicate.ne, value, zero)
        else:
            trace("value_type.width > target_type.width, truncating")
            value = arith.trunci(out=work_target_type, in_=value)
    elif work_value_type.width < work_target_type.width:
        trace("value_type.width < target_type.width, extending")
        # Default to unsigned extension for signless integers (common in GPU code)
        # Use signed extension only when explicitly marked as signed
        extend_op = arith.extsi if use_signed_extend else arith.extui
        value = extend_op(out=work_target_type, in_=value)
    elif work_value_type != work_target_type:
        trace("bitcasting to match intermediate target signedness")
        value = arith.bitcast(out=work_target_type, in_=value)

    if work_target_type != target_type:
        trace("restoring requested target signedness via bitcast")
        value = arith.bitcast(out=target_type, in_=value)

    return value


def convert(value, target_type, *, signed: bool = False):
    if getattr(value, "type", None) == target_type:
        return value
    return ensure_verifies(unverified_convert(value, target_type, signed=signed))


@singledispatch
def unverified_convert(value, target_type, *, signed: bool = False):
    raise NotImplementedError(f"Not implemented for type {type(value)}")


@unverified_convert.register
def convert_none(value: ir.NoneType, target_type: ir.NoneType, **_):
    if value != target_type:
        raise InternalCompilerError("Cannot convert NoneType to anything other than NoneType")
    return value


@unverified_convert.register(type)
@unverified_convert.register(MLIRDispatcher)
def opaque_data_model_convert(value, target_type: ir.NoneType, **_):
    """
    For types with an opaque data model, we defer the real lowering until later - we
    hopefully resolve this at compile time anyways.
    """
    return value


@unverified_convert.register
def number_class_convert(value: types.Type, target_type: types.functions.NumberClass, **_):
    return value


def _memory_spaces_match(lhs: ir.MemRefType, rhs: ir.MemRefType) -> bool:
    """
    `lhs.memory_space == rhs.memory_space` works when both memrefs _have_ a memory
    space, but when one of them does not, `memref.memory_space` returns None and
    attributes do not have an __eq__ overload for None, so an error is raised.
    That's why the weird XOR.
    """
    match lhs.memory_space, rhs.memory_space:
        case ir.Attribute() as a, ir.Attribute() as b if a == b:
            return True
        case None, ir.Attribute() | ir.Attribute(), None:
            return False
        case None, None:
            return True


@unverified_convert.register(ir.Value)
@unverified_convert.register(int)
@unverified_convert.register(float)
@unverified_convert.register(bool)
@unverified_convert.register(complex)
def unverified_basic_mlir_convert(
    value,
    target_type: ir.Type,
    *,
    signed: bool = False,
) -> ir.Value:
    from numba_cuda_mlir._mlir.dialects import (
        complex as complex_dialect,
        nvgpu,
        tensor,
        vector,
    )

    if isinstance(value, (int, float, bool, complex)):
        if isinstance(target_type, ir.IndexType):
            return index_of(int(value))
        return constant(value, target_type)
    value_type = value.type
    trace("value_type: %s, target_type: %s", value_type, target_type)
    match value_type, target_type:
        case ir.Type() as x, ir.Type() as y if x == y:
            trace("value_type == target_type, returning value")
            return value
        case (ir.IndexType(), ir.IntegerType()) | (ir.IntegerType(), ir.IndexType()):
            return arith.index_cast(out=target_type, in_=value)
        case (ir.IndexType(), ir.FloatType()) | (ir.FloatType(), ir.IndexType()):
            return convert(int_of(value, T.i64()), target_type)
        case ir.IntegerType() as a, ir.IntegerType() as b:
            return _convert_integer_to_integer(value, target_type, signed=signed)
        case ir.FloatType(), ir.FloatType() if (
            value_type.width == target_type.width and value_type != target_type
        ):
            # Same width but different float types (e.g. f16 <-> bf16)
            # Convert via f32 as intermediate
            trace("same-width float conversion via f32: %s -> %s", value_type, target_type)
            intermediate = arith.extf(out=T.f32(), in_=value)
            return arith.truncf(out=target_type, in_=intermediate)
        case ir.FloatType(), ir.FloatType() if value_type.width != target_type.width:
            if value_type.width > target_type.width:
                trace("value_type.width > target_type.width, truncating")
                return arith.truncf(out=target_type, in_=value)
            else:
                trace("value_type.width < target_type.width, extending")
                return arith.extf(out=target_type, in_=value)
        case ir.VectorType() as vt1, ir.VectorType() as vt2 if vt1.shape == vt2.shape:
            # We can use arith ops to convert vectors element-wise
            elem1 = vt1.element_type
            elem2 = vt2.element_type
            if isinstance(elem1, ir.FloatType) and isinstance(elem2, ir.FloatType):
                if elem1.width > elem2.width:
                    return arith.truncf(out=target_type, in_=value)
                else:
                    return arith.extf(out=target_type, in_=value)
            elif isinstance(elem1, ir.IntegerType) and isinstance(elem2, ir.IntegerType):
                if elem1.width > elem2.width:
                    return arith.trunci(out=target_type, in_=value)
                else:
                    return arith.extsi(out=target_type, in_=value)
            elif isinstance(elem1, ir.IntegerType) and isinstance(elem2, ir.FloatType):
                return (
                    arith.sitofp(out=target_type, in_=value)
                    if elem1.width > 1
                    else arith.uitofp(out=target_type, in_=value)
                )
            elif isinstance(elem1, ir.FloatType) and isinstance(elem2, ir.IntegerType):
                return (
                    arith.fptosi(out=target_type, in_=value)
                    if elem2.width > 1
                    else arith.fptoui(out=target_type, in_=value)
                )
            else:
                raise NotImplementedError(f"Vector conversion not implemented: {vt1} to {vt2}")
        case ir.IntegerType(), ir.FloatType():
            return (
                arith.sitofp(out=target_type, in_=value)
                if value_type.width > 1
                else arith.uitofp(out=target_type, in_=value)
            )
        case ir.BF16Type(), ir.IntegerType() if target_type.width == 16:
            # bf16 to int16/uint16: use bitcast to preserve bit pattern
            trace("bf16 -> i16 bitcast conversion")
            return arith.bitcast(out=target_type, in_=value)
        case ir.FloatType(), ir.IntegerType():
            return (
                arith.fptosi(out=target_type, in_=value)
                if target_type.width > 1
                else arith.fptoui(out=target_type, in_=value)
            )
        case ir.IntegerType(), ir.ComplexType():
            # Convert integer to complex: int -> float -> complex(float, 0)
            float_type = target_type.element_type
            float_val = (
                arith.sitofp(out=float_type, in_=value)
                if value_type.width > 1
                else arith.uitofp(out=float_type, in_=value)
            )
            zero = arith.constant(result=float_type, value=0.0)
            return complex_dialect.create_(complex=target_type, real=float_val, imaginary=zero)
        case ir.FloatType(), ir.ComplexType():
            # Convert float to complex: float -> complex(float, 0)
            float_type = target_type.element_type
            real_val = convert(value, float_type)
            zero = arith.constant(result=float_type, value=0.0)
            return complex_dialect.create_(complex=target_type, real=real_val, imaginary=zero)
        case ir.ComplexType(), ir.ComplexType():
            assert value_type.element_type != target_type.element_type, (
                "how did we get here? the types should compare-equal."
            )
            target_element_type = target_type.element_type
            real = complex_dialect.re(value)
            real = convert(real, target_element_type)
            imag = complex_dialect.im(value)
            imag = convert(imag, target_element_type)
            return complex_dialect.create_(complex=target_type, real=real, imaginary=imag)
        case ir.MemRefType() as mr, ptr_type if str(ptr_type) == "!llvm.ptr":
            idx = memref_data_pointer_as_index(value, mr.element_type)
            return convert(idx, target_type)
        case ptr_type, ir.IntegerType() if str(ptr_type) == "!llvm.ptr":
            ptrtoi = llvm.ptrtoint(res=T.i64(), arg=value)
            return convert(ptrtoi, target_type)
        case ptr_type, ir.IndexType() if str(ptr_type) == "!llvm.ptr":
            value = llvm.ptrtoint(res=T.i64(), arg=value)
            return convert(value, target_type)
        case ir.IntegerType(), ptr_type if str(ptr_type) == "!llvm.ptr":
            itoptr = convert(value, T.i64())
            return llvm.inttoptr(res=target_type, arg=itoptr)
        case ir.IndexType(), ptr_type if str(ptr_type) == "!llvm.ptr":
            value = convert(value, T.i64())
            return convert(value, target_type)
        case (
            ir.MemRefType() as value_type,
            ir.MemRefType() as target_type,
        ) if not _memory_spaces_match(value_type, target_type):
            memref_type_with_memory_space = ir.MemRefType.get(
                value_type.shape,
                value_type.element_type,
                value_type.layout,
                target_type.memory_space,
            )
            mr = memref.memory_space_cast(dest=memref_type_with_memory_space, source=value)
            return convert(mr, target_type)
        case ir.MemRefType() as a, ir.MemRefType() as b if not a.has_rank or not b.has_rank:
            raise NotImplementedError("Conversions between unranked memrefs")
        case ir.MemRefType() as a, ir.MemRefType() as b if a.rank == b.rank:
            return memref.cast(dest=target_type, source=value)
        case ir.MemRefType(), ir.MemRefType():
            value = memref_to_tensor(value)
            shape = [tensor.dim(value, index_of(i)) for i in range(target_type.rank)]
            shape = tensor.from_elements(T.tensor(target_type.rank, T.index()), shape)
            t_type = T.tensor(*target_type.shape, target_type.element_type)
            value = tensor.reshape(t_type, value, shape)
            value = tensor_to_memref(value)
            value = convert(
                value, target_type
            )  # should be compatible with a memref cast if not equal
            return value
        case nvgpu.TensorMapDescriptorType(), ir.Type() if str(target_type) == "!llvm.ptr":
            return builtin.unrealized_conversion_cast([llvm.PointerType.get()], [value])
        case ir.Type() as a, nvgpu.TensorMapDescriptorType() if str(a) == "!llvm.ptr":
            return builtin.unrealized_conversion_cast([target_type], [value])
        case ir.ComplexType(), ir.FloatType():
            return complex_dialect.re(value)
        case ir.FloatType(), ir.ComplexType():
            return complex_dialect.create_(
                complex=target_type,
                real=value,
                imaginary=float_of(0, target_type.element_type),
            )
        case ir.ComplexType() as ct, ir.VectorType() as vt if vt.shape == [2]:
            real = complex_dialect.re(value)
            imag = complex_dialect.im(value)
            if ct.element_type != vt.element_type:
                real = convert(real, vt.element_type)
                imag = convert(imag, vt.element_type)
            v = llvm.mlir_undef(vt)
            v = vector.insert(real, v, [], [0])
            v = vector.insert(imag, v, [], [1])
            return v
        case ir.VectorType() as vt, ir.ComplexType() as ct if vt.shape == [2]:
            real = vector.extract(value, [], [0])
            imag = vector.extract(value, [], [1])
            if vt.element_type != ct.element_type:
                real = convert(real, ct.element_type)
                imag = convert(imag, ct.element_type)
            return complex_dialect.create_(complex=ct, real=real, imaginary=imag)
        case ir.Type() as x, ir.Type() as y if x != y:
            raise NotImplementedError(
                f"Type cast not implemented: {x} to {y}. Should this be a bitcast? "
                "Please file an issue with your use case. Thank you!"
            )
        case _:
            raise NotImplementedError(f"NotImplemented converting {value_type} to {target_type}")


def index_of(value: ir.Value | int) -> ir.Value:
    match value:
        case int():
            return arith.constant(T.index(), value=value)
        case ir.Value():
            return convert(value, T.index())
        case _:
            raise NotImplementedError(f"Not implemented for type {type(value)}")


def int_of(value: ir.Value | int | float | bool, ty: ir.Type, *, signed: bool = False) -> ir.Value:
    match value:
        case int() | float() | bool():
            return arith.constant(ty, value=int(value))
        case ir.Value():
            return convert(value, ty, signed=signed)
        case _:
            raise NotImplementedError(f"Not implemented for type {type(value)}")


def float_of(value: ir.Value | float | int, ty: ir.Type) -> ir.Value:
    match value:
        case int() | float():
            return arith.constant(ty, value=float(value))
        case ir.Value():
            return convert(value, ty)
        case _:
            raise NotImplementedError(f"Not implemented for type {type(value)}")


def i32_of(value: ir.Value | int) -> ir.Value:
    return int_of(value, T.i32())


def i64_of(value: ir.Value | int) -> ir.Value:
    return int_of(value, T.i64())


def f32_of(value: ir.Value | float) -> ir.Value:
    return float_of(value, T.f32())


def f64_of(value: ir.Value | float) -> ir.Value:
    return float_of(value, T.f64())


def user_signature_to_external_abi_signature(
    signature: typing.Signature,
) -> typing.Signature:
    """
    The user's function must return an integer indicating if a Python exception occurred,
    and the return value is passed to the callee by pointer.

    See: https://nvidia.github.io/numba-cuda/user/cuda_ffi.html#Device-Function-ABI
    """
    return types.int32(types.CPointer(signature.return_type), *signature.args)


def coerce_to_shape_tuple(
    value: ir.Value | tuple[ir.Value | int, ...],
) -> tuple[ir.Value, ...]:
    def to_indices(values):
        return tuple(map(lambda x: index_of(x), values))

    match value:
        case tuple() if all(isinstance(x, ir.Value) for x in value):
            return to_indices(value)
        case tuple() if all(isinstance(x, int) for x in value):
            return to_indices(value)
        case ir.Value() if isinstance(value.type, ir.MemRefType):
            mr_type = value.type
            assert mr_type.has_rank and mr_type.rank == 1, "Value must be a 1-dimensional memref"
            assert mr_type.has_static_shape, "Shape of memref must be static"
            tuple_result = tuple(
                memref.load(value, [index_of(i)]) for i in range(mr_type.get_dim_size(0))
            )
            return to_indices(tuple_result)
        case _:
            raise NotImplementedError(f"Not implemented for type {type(value)}")


class DeferredLowering:
    """
    Class representing a method call that defers lowering to a later stage.
    When a method's attribute is retrieved from an object, we must
    return a function ready for lowering, but the lowering function
    may require access to the object itself. In this case, we capture
    the object and any additional required context and return a function
    that can be called later to perform the lowering.
    """

    @abstractmethod
    def __call__(self, builder, target, args, kwargs): ...


class DeferredMethodCall(DeferredLowering):
    def __init__(self, _self, lowering_function: AnyCallable[PS]):
        self._self = _self
        self.lowering_function = lowering_function

    def __call__(self, builder, target, args, kwargs):
        return self.lowering_function(builder, target, [self._self] + args, kwargs)


@dataclass
class RangeObject:
    """
    Holds the state of a range object.
    """

    @staticmethod
    def _unify_integer_types(*values: ir.Value) -> ir.Type:
        max_width = 0
        for val in values:
            match val.type:
                case ir.IntegerType():
                    max_width = max(max_width, val.type.width)
                case ir.IndexType():
                    max_width = max(max_width, 64)
                case _:
                    raise InternalCompilerError(f"Unsupported integer type: {val.type}")
        return ir.IntegerType.get_signless(max_width)

    def __init__(self, lower, start: ir.Value, stop: ir.Value, step: ir.Value):
        element_type = self._unify_integer_types(start, stop, step)

        with lower.alloca_insertion_point():
            self._memref = memref.alloca(T.memref(5, element_type), [], [])

        start_val = convert(start, element_type, signed=True)
        stop_val = convert(stop, element_type, signed=True)
        step_val = convert(step, element_type, signed=True)
        zero = arith.constant(element_type, 0)
        one = arith.constant(element_type, 1)

        # Store start/stop/step
        memref.store(start_val, self._memref, [index_of(0)])
        memref.store(stop_val, self._memref, [index_of(1)])
        memref.store(step_val, self._memref, [index_of(2)])

        # Step must not be zero - add runtime check if error checking is enabled
        error_memref = lower._get_or_create_error_global()
        if error_memref is not None:
            from numba_cuda_mlir.mlir_lowering import KERNEL_ERROR_CODES

            step_is_zero = arith.cmpi(predicate=arith.CmpIPredicate.eq, lhs=step_val, rhs=zero)
            with scf.if_ctx_manager(step_is_zero):
                set_error_code_if_zero(error_memref, KERNEL_ERROR_CODES[ZeroDivisionError])
                scf.yield_([])

        # Compute trip-count in i64 to avoid issues with mixed signed/unsigned
        # semantics on narrow integer types, then truncate back.
        i64 = T.i64()
        start_i64 = convert(start_val, i64, signed=True)
        stop_i64 = convert(stop_val, i64, signed=True)
        step_i64 = convert(step_val, i64, signed=True)
        zero_i64 = arith.constant(i64, 0)
        one_i64 = arith.constant(i64, 1)

        diff = arith.subi(lhs=stop_i64, rhs=start_i64)
        neg_diff = arith.subi(lhs=zero_i64, rhs=diff)
        neg_step = arith.subi(lhs=zero_i64, rhs=step_i64)
        step_positive = arith.cmpi(arith.CmpIPredicate.sgt, lhs=step_i64, rhs=zero_i64)

        # Positive step: ceildiv(diff, step) if diff > 0 else 0
        pos_numerator = arith.addi(lhs=diff, rhs=arith.subi(lhs=step_i64, rhs=one_i64))
        pos_count = arith.divui(lhs=pos_numerator, rhs=step_i64)
        diff_positive = arith.cmpi(arith.CmpIPredicate.sgt, lhs=diff, rhs=zero_i64)
        pos_result = arith.select(diff_positive, pos_count, zero_i64)

        # Negative step: ceildiv(neg_diff, neg_step) if neg_diff > 0 else 0
        neg_numerator = arith.addi(lhs=neg_diff, rhs=arith.subi(lhs=neg_step, rhs=one_i64))
        neg_count = arith.divui(lhs=neg_numerator, rhs=neg_step)
        neg_diff_positive = arith.cmpi(arith.CmpIPredicate.sgt, lhs=neg_diff, rhs=zero_i64)
        neg_result = arith.select(neg_diff_positive, neg_count, zero_i64)

        count = arith.select(step_positive, pos_result, neg_result)
        count = convert(count, element_type, signed=True)
        memref.store(count, self._memref, [index_of(3)])
        memref.store(start_val, self._memref, [index_of(4)])

    def get_mlir_type(self) -> ir.Type:
        return self._memref.type

    @property
    def start(self) -> ir.Value:
        return memref.load(self._memref, [index_of(0)])

    @property
    def stop(self) -> ir.Value:
        return memref.load(self._memref, [index_of(1)])

    @property
    def step(self) -> ir.Value:
        return memref.load(self._memref, [index_of(2)])

    @property
    def count(self) -> ir.Value:
        return memref.load(self._memref, [index_of(3)])

    @property
    def iter(self) -> ir.Value:
        return memref.load(self._memref, [index_of(4)])

    def __next__(self) -> ir.Value:
        raise NotImplementedError()

    def next(self) -> ir.Value:
        """
        Compute (value, is_valid) and update internal state.
        """
        int_type = self._memref.type.element_type
        zero = int_of(0, int_type)
        one = int_of(1, int_type)

        current_count = self.count
        is_valid = arith.cmpi(predicate=arith.CmpIPredicate.sgt, lhs=current_count, rhs=zero)
        current_value = self.iter

        # Compute updated state (predicated without control flow)
        decremented = arith.subi(lhs=current_count, rhs=one)
        next_value = arith.addi(lhs=current_value, rhs=self.step)
        updated_count = arith.select(
            condition=is_valid, true_value=decremented, false_value=current_count
        )
        updated_iter = arith.select(
            condition=is_valid, true_value=next_value, false_value=current_value
        )

        # Store updates
        memref.store(updated_count, self._memref, [index_of(3)])
        memref.store(updated_iter, self._memref, [index_of(4)])

        result_type = T.memref(2, int_type)
        result = memref.alloca(memref=result_type, dynamic_sizes=[], symbol_operands=[])
        memref.store(current_value, result, [index_of(0)])
        memref.store(convert(is_valid, int_type), result, [index_of(1)])
        return result


class IterResult:
    """Holds (value, is_valid) result from iterator next() operation."""

    def __init__(self, value: ir.Value, is_valid: ir.Value):
        self.value = value
        self.is_valid = is_valid


def _zero_value_for_type(mlir_type):
    if isinstance(mlir_type, (ir.IntegerType, ir.IndexType)):
        return arith.constant(result=mlir_type, value=0)
    if isinstance(mlir_type, ir.FloatType):
        return arith.constant(result=mlir_type, value=0.0)
    if isinstance(mlir_type, ir.ComplexType):
        zero = _zero_value_for_type(mlir_type.element_type)
        return complex_dialect.create_(complex=mlir_type, real=zero, imaginary=zero)
    if isinstance(mlir_type, ir.VectorType):
        zero = _zero_value_for_type(mlir_type.element_type)
        return vector_dialect.broadcast(mlir_type, zero)
    if str(mlir_type).startswith("!llvm."):
        return llvm.mlir_zero(res=mlir_type)
    raise InternalCompilerError(f"Cannot materialize iterator dummy value for {mlir_type}.")


@dataclass
class UniTupleIterObject:
    """
    Holds the state of a UniTuple iterator object.

    Supports two storage modes:
    - memref (for builtin scalar element types like i32, f64, index)
    - llvm alloca (for LLVM dialect types like !llvm.ptr, !llvm.struct)

    The storage mode is chosen automatically based on the element type.
    """

    def __init__(
        self,
        lower,
        tuple_storage: ir.Value,
        count: int,
        element_type: ir.Type,
        *,
        uses_llvm_storage: bool = False,
    ):
        self._tuple_storage = tuple_storage
        self._count = count
        self._element_type = element_type
        self._uses_llvm = uses_llvm_storage
        with lower.alloca_insertion_point():
            self._index_memref = memref.alloca(T.memref(1, T.i64()), [], [])
        memref.store(int_of(0, ty=T.i64()), self._index_memref, [index_of(0)])

    @property
    def index(self) -> ir.Value:
        return memref.load(self._index_memref, [index_of(0)])

    def next(self) -> IterResult:
        """
        Compute (value, is_valid) and update internal state.
        Returns an IterResult containing the value and validity flag.
        """
        one = int_of(1, ty=T.i64())
        count = int_of(self._count, ty=T.i64())

        current_index = self.index
        is_valid = arith.cmpi(predicate=arith.CmpIPredicate.slt, lhs=current_index, rhs=count)

        load_if_valid = scf.IfOp(is_valid, results_=[self._element_type], has_else=True)
        with ir.InsertionPoint(load_if_valid.then_block):
            if self._uses_llvm:
                elem_ptr = llvm.getelementptr(
                    llvm.PointerType.get(),
                    self._tuple_storage,
                    [current_index],
                    [GEP_DYNAMIC_INDEX],
                    self._element_type,
                    None,
                )
                current_value = llvm.load(res=self._element_type, addr=elem_ptr)
            else:
                index_as_index = arith.index_cast(out=ir.IndexType.get(), in_=current_index)
                current_value = memref.load(self._tuple_storage, [index_as_index])
            scf.yield_([current_value])
        with ir.InsertionPoint(load_if_valid.else_block):
            scf.yield_([_zero_value_for_type(self._element_type)])
        current_value = load_if_valid.results[0]

        next_index = arith.addi(lhs=current_index, rhs=one)
        updated_index = arith.select(
            condition=is_valid, true_value=next_index, false_value=current_index
        )
        memref.store(updated_index, self._index_memref, [index_of(0)])

        return IterResult(current_value, is_valid)


@dataclass
class ArrayIterObject:
    """
    Holds the state of an array iterator object.
    """

    def __init__(self, lower, array: ir.Value, element_type: ir.Type, length: ir.Value):
        self._array = array
        self._element_type = element_type
        with lower.alloca_insertion_point():
            self._index_memref = memref.alloca(T.memref(1, T.i64()), [], [])
            self._length_memref = memref.alloca(T.memref(1, T.i64()), [], [])
        memref.store(int_of(0, ty=T.i64()), self._index_memref, [index_of(0)])
        memref.store(int_of(length, ty=T.i64()), self._length_memref, [index_of(0)])

    @property
    def index(self) -> ir.Value:
        return memref.load(self._index_memref, [index_of(0)])

    @property
    def length(self) -> ir.Value:
        return memref.load(self._length_memref, [index_of(0)])

    def next(self) -> IterResult:
        """
        Compute (value, is_valid) and update internal state.
        Returns an IterResult containing the value and validity flag.
        """
        one = int_of(1, ty=T.i64())

        current_index = self.index
        length = self.length

        is_valid = arith.cmpi(predicate=arith.CmpIPredicate.slt, lhs=current_index, rhs=length)

        load_if_valid = scf.IfOp(is_valid, results_=[self._element_type], has_else=True)
        with ir.InsertionPoint(load_if_valid.then_block):
            index_as_index = arith.index_cast(out=ir.IndexType.get(), in_=current_index)
            current_value = memref.load(self._array, [index_as_index])
            scf.yield_([current_value])
        with ir.InsertionPoint(load_if_valid.else_block):
            scf.yield_([_zero_value_for_type(self._element_type)])
        current_value = load_if_valid.results[0]

        next_index = arith.addi(lhs=current_index, rhs=one)
        updated_index = arith.select(
            condition=is_valid, true_value=next_index, false_value=current_index
        )
        memref.store(updated_index, self._index_memref, [index_of(0)])

        return IterResult(current_value, is_valid)


def _make_zero_d_subview(array_mr: ir.Value, multi_idx: list[ir.Value]) -> ir.Value:
    """Build a rank-0 memref view at `multi_idx` of `array_mr`.

    All elements of multi_idx must be IndexType values; their count must equal
    the rank of array_mr.  The sizes and strides are passed as the integer 1
    (not as SSA values) so MLIR sees them as static and can rank-reduce the
    result to rank 0.
    """
    source_type: ir.MemRefType = array_mr.type
    dyn_stride = ir.ShapedType.get_dynamic_stride_or_offset()
    result_type = ir.MemRefType.get(
        shape=[],
        element_type=source_type.element_type,
        layout=ir.StridedLayoutAttr.get(dyn_stride, []),
        memory_space=source_type.memory_space,
    )
    rank = source_type.rank
    return memref.subview(
        array_mr,
        offsets=multi_idx,
        sizes=[1] * rank,
        strides=[1] * rank,
        result_type=result_type,
    )


@dataclass
class NdIterIterObject:
    """
    Iterator state for ``np.nditer`` over one or more arrays.

    Each call to ``next()`` yields either a rank-0 memref view of the current
    element (single-array case) or a tuple of rank-0 views (multi-array case).
    Iteration proceeds in C order over the broadcast shape.
    """

    def __init__(
        self,
        lower,
        arrays: list[ir.Value],
        ndim: int,
        broadcast_shape: list[ir.Value] | None = None,
    ):
        # arrays may have differing rank when broadcasting; broadcast_shape
        # is the iteration shape (length == ndim, all IndexType values).
        # When all arrays are the same shape, callers may pass None and the
        # shape is taken from the first array.
        self._arrays = list(arrays)
        self._ndim = ndim
        if broadcast_shape is None:
            broadcast_shape = [memref.dim(arrays[0], index_of(d)) for d in range(ndim)]
        self._shape = list(broadcast_shape)
        with lower.alloca_insertion_point():
            self._index_memref = memref.alloca(T.memref(1, T.i64()), [], [])
        memref.store(int_of(0, ty=T.i64()), self._index_memref, [index_of(0)])
        # Total element count = product of broadcast shape.
        if ndim == 0:
            size_i64 = int_of(1, ty=T.i64())
        else:
            size_idx = self._shape[0]
            for d in range(1, ndim):
                size_idx = arith.muli(size_idx, self._shape[d])
            size_i64 = arith.index_cast(out=T.i64(), in_=size_idx)
        with lower.alloca_insertion_point():
            self._size_memref = memref.alloca(T.memref(1, T.i64()), [], [])
        memref.store(size_i64, self._size_memref, [index_of(0)])

    @property
    def index(self) -> ir.Value:
        return memref.load(self._index_memref, [index_of(0)])

    @property
    def size(self) -> ir.Value:
        return memref.load(self._size_memref, [index_of(0)])

    def _unravel_c_order(self, linear_idx_i64: ir.Value) -> list[ir.Value]:
        """Convert linear index (i64) to a list of IndexType per-dim indices, C order."""
        if self._ndim == 0:
            return []
        if self._ndim == 1:
            return [index_of(linear_idx_i64)]
        per_dim_i64 = [None] * self._ndim
        remaining = linear_idx_i64
        for d in reversed(range(self._ndim)):
            dim_size_i64 = arith.index_cast(out=T.i64(), in_=self._shape[d])
            per_dim_i64[d] = arith.remsi(lhs=remaining, rhs=dim_size_i64)
            remaining = arith.divsi(lhs=remaining, rhs=dim_size_i64)
        return [index_of(v) for v in per_dim_i64]

    def next(self) -> IterResult:
        one_i64 = int_of(1, ty=T.i64())
        current_index = self.index
        size = self.size
        is_valid = arith.cmpi(predicate=arith.CmpIPredicate.slt, lhs=current_index, rhs=size)

        multi_idx = self._unravel_c_order(current_index)
        views = [_make_zero_d_subview(arr, multi_idx) for arr in self._arrays]

        next_index = arith.addi(lhs=current_index, rhs=one_i64)
        updated_index = arith.select(
            condition=is_valid, true_value=next_index, false_value=current_index
        )
        memref.store(updated_index, self._index_memref, [index_of(0)])

        if len(views) == 1:
            return IterResult(views[0], is_valid)
        return IterResult(tuple(views), is_valid)


def _types_match(ty1, ty2, exact=False):
    if exact:
        return ty1 == ty2
    match ty1, ty2:
        case a, b if a == b:
            return True
        case ir.MemRefType() as mr1, ir.MemRefType() as mr2:
            matches = mr1.element_type == mr2.element_type
            matches &= mr1.rank == mr2.rank
            matches &= _memory_spaces_match(mr1, mr2)
            return matches
        case (ir.IndexType(), ir.IntegerType() as i) | (
            ir.IntegerType() as i,
            ir.IndexType(),
        ):
            return i.width == 64
        case _:
            return False


def get_func_type(op: func.FuncOp | gpu.GPUFuncOp) -> ir.FunctionType:
    """Get the FunctionType from a func.FuncOp or gpu.GPUFuncOp."""
    if isinstance(op, func.FuncOp):
        return op.type
    else:
        return op.function_type.value


def _func_types_match(ty1: ir.FunctionType, ty2: ir.FunctionType, exact=False):
    if exact:
        return ty1 == ty2
    if len(ty1.inputs) != len(ty2.inputs):
        return False
    if len(ty1.results) != len(ty2.results):
        return False
    for i in range(len(ty1.inputs)):
        if not _types_match(ty1.inputs[i], ty2.inputs[i], exact):
            return False
    for i in range(len(ty1.results)):
        if not _types_match(ty1.results[i], ty2.results[i], exact):
            return False
    return True


def lookup_callee_in_module(
    name: str,
    mlir_type: ir.FunctionType,
    module: ir.Module | gpu.GPUModuleOp,
    exact=False,
) -> func.FuncOp | gpu.GPUFuncOp | None:
    trace("looking up callee %s", name)
    name = ir.StringAttr.get(name)
    body = module.regions[0].blocks[0]
    func_ops = list(filter(lambda x: isinstance(x, (func.FuncOp, gpu.GPUFuncOp)), body))
    matches = list(
        filter(
            lambda x: x.name == name and _func_types_match(get_func_type(x), mlir_type, exact),
            func_ops,
        )
    )
    if len(matches) == 1:
        return matches[0]
    elif len(matches) > 1:
        raise ValueError(
            f"Multiple functions with name {name} and type {mlir_type} found in module"
        )
    return None


def declare_external_function(
    name: str, mlir_type: ir.FunctionType, ip: ir.InsertionPoint
) -> func.FuncOp:
    with ip:
        func_op = func.FuncOp(name=name, type=mlir_type, visibility="private")
    return func_op


def get_or_insert_function(
    name: str, mlir_type: ir.FunctionType, module: ir.Module | gpu.GPUModuleOp
) -> func.FuncOp | gpu.GPUFuncOp:
    if callee := lookup_callee_in_module(name, mlir_type, module):
        return callee
    body = module.regions[0].blocks[0]
    return declare_external_function(name, mlir_type, ir.InsertionPoint(body))


def constant(
    value: ir.Value | int | float | bool | complex | np.number, ty: ir.Type | None
) -> ir.Value:
    """
    Creating an arith.constant for a float type and passing an integer value
    will cause a segfault in MLIR(!!!) so this utility converts the value
    to the appropriate type. This should really be handled in MLIR.
    TODO: send patch
    """
    from numba_cuda_mlir._mlir.dialects import complex as complex_dialect

    match value, ty:
        # First, if the type is None, we infer the MLIR type.
        # This is useful for converting literals to the appropriate type.
        case ir.Value(), t if t is None:
            return value
        case int(), ir.IntegerType():
            return arith.constant(ty, value=value)
        case int(), ir.IndexType():
            return arith.constant(ty, value=value)
        case float(), ir.FloatType():
            return arith.constant(ty, value=value)
        case bool(), None:
            return arith.constant(T.bool(), value=value)
        case complex(), None:
            element_type = T.f64()
            complex_type = T.complex(element_type)
            real_attr = ir.FloatAttr.get(element_type, value.real)
            imag_attr = ir.FloatAttr.get(element_type, value.imag)
            const_attr = ir.ArrayAttr.get([real_attr, imag_attr])
            return complex_dialect.constant(complex=complex_type, value=const_attr)
        case ir.Value(), _:
            return convert(value, ty)
        case _, ir.IntegerType():
            return arith.constant(ty, value=int(value))
        case _, ir.IndexType():
            return arith.constant(ty, value=int(value))
        case _, ir.FloatType():
            return arith.constant(ty, value=float(value))
        case _, ir.ComplexType():
            element_type = ty.element_type
            cplx = complex(value)
            real_attr = ir.FloatAttr.get(element_type, cplx.real)
            imag_attr = ir.FloatAttr.get(element_type, cplx.imag)
            const_attr = ir.ArrayAttr.get([real_attr, imag_attr])
            return complex_dialect.constant(complex=ty, value=const_attr)
        case _:
            # If we get here with a type that doesn't support constants,
            # provide a helpful error message
            import traceback

            stack = "".join(traceback.format_stack())
            raise NotImplementedError(
                f"constant() not implemented for value={value} (type: {type(value).__name__}), "
                f"ty={ty} (type class: {type(ty).__name__ if ty else 'None'})\n"
                f"Stack trace:\n{stack}"
            )


def dims_of_tensor_shape(tens: ir.Value) -> list[ir.Value]:
    from numba_cuda_mlir._mlir.dialects import shape

    ty = tens.type
    sh = shape.shape_of(tens)
    return [shape.get_extent(sh, index_of(i)) for i in range(ty.rank)]


def broadcast_shapes_for_binary_op(
    a: ir.Value, b: ir.Value, builder=None
) -> tuple[ir.Value, ir.Value]:
    from numba_cuda_mlir._mlir.dialects import shape, tensor

    if isinstance(a.type, ir.MemRefType):
        a = memref_to_tensor(a)
    if isinstance(b.type, ir.MemRefType):
        b = memref_to_tensor(b)
    match a.type, b.type:
        case ir.RankedTensorType(), ir.RankedTensorType():
            if a.type.rank == b.type.rank:
                return a, b
            sa, sb = shape.shape_of(a), shape.shape_of(b)
            is_broadcastable = shape.is_broadcastable([sa, sb])
            if builder is not None:
                error_memref = builder._get_or_create_error_global()
                if error_memref is not None:
                    from numba_cuda_mlir.mlir_lowering import KERNEL_ERROR_CODES

                    not_broadcastable = arith.xori(
                        is_broadcastable, arith.constant(result=T.i(1), value=1)
                    )
                    with scf.if_ctx_manager(not_broadcastable):
                        set_error_code_if_zero(error_memref, KERNEL_ERROR_CODES[ValueError])
                        scf.yield_([])
            sh = shape.broadcast(shapes=[sb, sa], result=sa.type)
            a, b = tensor.reshape(a, sh), tensor.reshape(b, sh)
            return a, b
        case ir.RankedTensorType() as t, ir.IntegerType() | ir.FloatType():
            tensor_value, scalar_value = a, b
            scalar_value = constant(scalar_value, t.element_type)
            splatted = tensor.splat(t, scalar_value, dims_of_tensor_shape(tensor_value))
            return tensor_value, splatted
        case ir.IntegerType() | ir.FloatType(), ir.RankedTensorType() as t:
            scalar_value, tensor_value = a, b
            scalar_value = constant(scalar_value, t.element_type)
            splatted = tensor.splat(t, scalar_value, dims_of_tensor_shape(tensor_value))
            return splatted, tensor_value
        case ir.IntegerType() | ir.FloatType(), ir.IntegerType() | ir.FloatType():
            return a, b
        case _:
            raise NotImplementedError(f"types {a.type} and {b.type}")


def _simple_scalar_conversion_op(src: ir.Type, dst: ir.Type):
    match src, dst:
        case x, y if x == y:
            return lambda dst, x: x
        case ir.FloatType(), ir.FloatType():
            return arith.truncf if src.width > dst.width else arith.extf
        case ir.IntegerType(), ir.IntegerType():
            return arith.trunci if src.width > dst.width else arith.extsi
        case ir.FloatType(), ir.IntegerType():
            return arith.fptosi
        case ir.IntegerType(), ir.FloatType():
            return arith.sitofp if src.width > 1 else arith.uitofp
        case _:
            raise NotImplementedError(f"Not implemented for types {src} and {dst}")


def simple_scalar_conversion_op(src: ir.Type, dst: ir.Type):
    op = _simple_scalar_conversion_op(src, dst)
    return functools.partial(op, dst)


def expensive_coerce_tensor_type(a: ir.Value, target_element_type: ir.Type) -> ir.Value:
    from numba_cuda_mlir._mlir.dialects import linalg, tensor

    if a.type.element_type == target_element_type:
        return a
    output_type = T.tensor(*a.type.shape, target_element_type)
    src, dst = a.type.element_type, target_element_type
    op = simple_scalar_conversion_op(src, dst)
    init = tensor.empty(sizes=dims_of_tensor_shape(a), element_type=target_element_type)
    result = linalg.MapOp(
        result=[output_type],
        inputs=[a],
        init=init,
    )
    block = result.mapper.blocks.append(a.type.element_type, target_element_type)
    with ir.InsertionPoint(block):
        linalg.yield_([op(block.arguments[0])])
    return result.result[0]


def try_extract_constant(
    value: ir.Value | ir.OpResult | int | float | bool | None,
) -> int | float | bool | None:
    if isinstance(value, np.generic):
        # numpy scalars (including np.bool_, which is not an np.number)
        # can reach here as frozen closure or global constants.
        value = value.item()
    match value:
        case int() | float() | bool() | None:
            return value
    value = ir.Value(value)
    owner = value.owner
    if isinstance(owner, ir.Block):
        return None

    # Get the OpView - owner might be ir.Operation or already an OpView
    if isinstance(owner, ir.Operation):
        opview = owner.opview
    else:
        # owner is already an OpView (e.g., arith.ConstantOp)
        opview = owner

    match opview:
        case arith.ConstantOp(value=v):
            return v.value
        case (
            arith.IndexCastOp()
            | arith.FPToSIOp()
            | arith.FPToUIOp()
            | arith.SIToFPOp()
            | arith.UIToFPOp()
        ):
            return try_extract_constant(opview.operands[0])
        case _:
            return None


def is_nonelike(ty):
    """returns if 'ty' is none"""
    return ty is None or isinstance(ty, types.NoneType) or isinstance(ty, types.Omitted)



# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""Whole-function extension planning after device-function inlining."""

import operator
from threading import RLock

from numba_cuda_mlir._launch_config import (
    _LAUNCH_CONFIG_TRACKER_METADATA_KEY,
    _LaunchConfigTracker,
)
from numba_cuda_mlir.numba_cuda.core import postproc
from numba_cuda_mlir.numba_cuda.core.ir_utils import build_definitions, simplify_CFG


_REQUIRED_DYNAMIC_SHARED_MEMORY_KEY = "required_dynamic_shared_memory"


class _RequireLaunchConfig(RuntimeError):
    """Report that no configured launch metadata is available."""


class WholeFunctionPlanner:
    """Base class for whole-function extension planners.

    Planners run after device-function inlining and before type inference. They
    may inspect and rewrite any block in ``state.func_ir`` and must return a
    Boolean indicating whether they changed the IR.
    """

    def __init__(self, state):
        self.state = state

    @property
    def is_device_function(self) -> bool:
        """Whether the current compilation is for a device function."""

        targetoptions = self.state.metadata.get("targetoptions", {})
        return bool(targetoptions.get("device", False))

    def run(self) -> bool:
        """Inspect or rewrite the current function and report whether it changed."""

        raise NotImplementedError


class _WholeFunctionPlannerRegistry:
    def __init__(self):
        self._planners = []
        self._lock = RLock()

    def register(self, planner_cls):
        """Register a planner class and return it for decorator use."""

        if not isinstance(planner_cls, type) or not issubclass(planner_cls, WholeFunctionPlanner):
            raise TypeError(f"{planner_cls!r} is not a WholeFunctionPlanner subclass")
        with self._lock:
            if planner_cls not in self._planners:
                self._planners.append(planner_cls)
        return planner_cls

    @property
    def has_planners(self) -> bool:
        with self._lock:
            return bool(self._planners)

    def apply(self, state) -> bool:
        """Run each planner once with coherent IR and repair every mutation."""

        modified = False
        with self._lock:
            planners = tuple(self._planners)
        if planners:
            # Inlining and CFG simplification can leave definitions and
            # postprocessing analysis describing blocks that no longer exist.
            # Repair once so the first inspection-only planner sees the same
            # coherent IR promised to every planner after a mutation.
            self._repair_ir(state.func_ir)
        for planner_cls in planners:
            result = planner_cls(state).run()
            if not isinstance(result, bool):
                raise TypeError(
                    f"{planner_cls.__name__}.run() must return bool, got {type(result)!r}"
                )
            if result:
                self._repair_ir(state.func_ir)
                modified = True
        return modified

    @staticmethod
    def _repair_ir(func_ir) -> None:
        func_ir.blocks = simplify_CFG(func_ir.blocks)
        func_ir._reset_analysis_variables()
        postproc.PostProcessor(func_ir).run()
        func_ir._definitions = build_definitions(func_ir.blocks)
        for block in func_ir.blocks.values():
            block.verify()


_planner_registry = _WholeFunctionPlannerRegistry()


def require_launch_config(state) -> dict:
    """Return normalized launch metadata for the current compiler attempt."""

    metadata = getattr(state, "metadata", None)
    if not isinstance(metadata, dict):
        raise TypeError("compiler state metadata must be a dict")
    targetoptions = metadata.get("targetoptions")
    if not isinstance(targetoptions, dict):
        raise TypeError("compiler state metadata must contain targetoptions")
    launch_config = targetoptions.get("__launch_config__")
    if (
        not isinstance(launch_config, dict)
        or "grid" not in launch_config
        or "block" not in launch_config
    ):
        launch_config_tracker = metadata.get(_LAUNCH_CONFIG_TRACKER_METADATA_KEY)
        if not isinstance(launch_config_tracker, _LaunchConfigTracker):
            raise _RequireLaunchConfig(
                "whole-function planner requires metadata from a configured kernel launch"
            )
        launch_config = launch_config_tracker.require()
        targetoptions["__launch_config__"] = launch_config
    return launch_config


def set_required_dynamic_shared_memory(state, size_in_bytes) -> None:
    """Record the minimum dynamic shared memory required by this compile.

    Repeated calls retain the largest requirement. The value affects the
    eventual launch without changing the configured launch specialization key.
    """

    if isinstance(size_in_bytes, bool):
        raise TypeError("required dynamic shared memory must be an integer")
    try:
        size_in_bytes = operator.index(size_in_bytes)
    except TypeError:
        raise TypeError("required dynamic shared memory must be an integer") from None
    if size_in_bytes < 0:
        raise ValueError("required dynamic shared memory cannot be negative")

    metadata = getattr(state, "metadata", None)
    if not isinstance(metadata, dict):
        raise TypeError("compiler state metadata must be a dict")
    previous = metadata.get(_REQUIRED_DYNAMIC_SHARED_MEMORY_KEY, 0)
    metadata[_REQUIRED_DYNAMIC_SHARED_MEMORY_KEY] = max(previous, size_in_bytes)


def register_planner(planner_cls):
    """Register a whole-function planner class.

    Register planners before compiling any dispatcher that needs them.
    Registration does not invalidate existing in-memory overloads.
    """

    return _planner_registry.register(planner_cls)


__all__ = [
    "WholeFunctionPlanner",
    "register_planner",
    "require_launch_config",
    "set_required_dynamic_shared_memory",
]



# Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
# See https://llvm.org/LICENSE.txt for license information.
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
import functools
import inspect
import io
import types
from typing import Dict

from numba_cuda_mlir.numba_cuda.cloudpickle import cloudpickle

from numba_cuda_mlir._mlir.ir import Type


def bind(func, instance, as_name=None):
    if as_name is None:
        as_name = func.__name__
    bound_method = func.__get__(instance, instance.__class__)
    setattr(instance, as_name, bound_method)
    return bound_method


class _empty_cell_value:
    """Sentinel for empty closures."""

    @classmethod
    def __reduce__(cls):
        return cls.__name__


_empty_cell_value = _empty_cell_value()


# based on https://github.com/cloudpipe/cloudpickle/blob/f111f7ab6d302e9b1e2a568d0e4c574895db6a6e/cloudpickle/cloudpickle.py#L513
def make_empty_cell():
    if False:
        # trick the compiler into creating an empty cell in our lambda
        cell = None
        raise AssertionError("this route should not be executed")

    return (lambda: cell).__closure__[0]


def make_cell(value=_empty_cell_value):
    cell = make_empty_cell()
    if value is not _empty_cell_value:
        cell.cell_contents = value
    return cell


def unpickle_mlir_type(v):
    return Type.parse(v)


class MLIRTypePickler(cloudpickle.Pickler):
    def reducer_override(self, obj):
        if isinstance(obj, Type):
            return unpickle_mlir_type, (str(obj),)
        return super().reducer_override(obj)


def copy_object(obj):
    # see https://github.com/cloudpipe/cloudpickle/blob/f111f7ab6d302e9b1e2a568d0e4c574895db6a6e/cloudpickle/cloudpickle.py#L813
    # for how this trick is accomplished (dill and pickle both fail to pickle eg generic typevars)
    with io.BytesIO() as file:
        cp = MLIRTypePickler(file)
        cp.dump(obj)
        obj = cloudpickle.loads(file.getvalue())
    return obj


# Based on http://stackoverflow.com/a/6528148/190597 (Glenn Maynard);
# potentially more complete approach https://stackoverflow.com/a/56901529/9045206
def copy_func(f, new_closure: Dict = None):
    if new_closure is not None:
        # closure vars go into co_freevars
        code = f.__code__.replace(co_freevars=tuple(new_closure.keys()))
        # closure is a tuple of cells
        closure = tuple(
            make_cell(v) if not isinstance(v, types.CellType) else v for v in new_closure.values()
        )
    else:
        closure = copy_object(f.__closure__)
        code = f.__code__

    g = types.FunctionType(
        code=code,
        globals=f.__globals__,
        name=f.__name__,
        argdefs=f.__defaults__,
        closure=closure,
    )
    g.__kwdefaults__ = f.__kwdefaults__
    g.__dict__.update(f.__dict__)
    g = functools.update_wrapper(g, f)

    if inspect.ismethod(f):
        g = bind(g, f.__self__)
    return g



# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
from numba_cuda_mlir.descriptor import MLIRDispatcher
from numba_cuda_mlir.lowering_utilities.type_conversions import to_numba_type
import inspect
from typing import Callable, Any, TypeVar
from numba_cuda_mlir import types, typing
from functools import lru_cache
from numba_cuda_mlir.numba_cuda.typing.templates import (
    ConcreteTemplate,
    AttributeTemplate,
    Registry,
)
from numba_cuda_mlir.numba_cuda.typing.typeof import typeof_impl
from numba_cuda_mlir.numba_cuda.core.imputils import lower_builtin
from pathlib import Path
from numba_cuda_mlir.numba_cuda.codegen import ExternalCodeLibrary
from numba_cuda_mlir.numba_cuda.compiler import sigutils
from numba_cuda_mlir.descriptor import mlir_target
from numba_cuda_mlir.numba_cuda.core import funcdesc
from numba_cuda_mlir.logging import trace
from numba_cuda_mlir.typing.externals import (
    ExternMLIRLibrary,
    ExternMLIRLibraryFunction,
)

T = TypeVar("T")


class CUFunc:
    """
    Wrapper around a CUDA function handle providing configuration methods.

    This provides the same interface as numba-cuda's cufunc objects for
    configuring cache preferences and shared memory carveout.
    """

    def __init__(self, handle):
        """
        Initialize with a CUfunction handle.

        Args:
            handle: A CUfunction handle from cuModuleGetFunction
        """
        self._handle = handle

    @property
    def handle(self):
        return self._handle

    def cache_config(self, prefer_shared=False, prefer_cache=False, prefer_equal=False):
        """
        Set the cache configuration preference for this function.

        Args:
            prefer_shared: Prefer more shared memory over L1 cache
            prefer_cache: Prefer more L1 cache over shared memory
            prefer_equal: Prefer equal shared memory and L1 cache

        If no preference is specified, uses the default (no preference).
        """
        from cuda.bindings import driver

        # Determine cache config based on arguments
        if prefer_shared:
            config = driver.CUfunc_cache.CU_FUNC_CACHE_PREFER_SHARED
        elif prefer_cache:
            config = driver.CUfunc_cache.CU_FUNC_CACHE_PREFER_L1
        elif prefer_equal:
            config = driver.CUfunc_cache.CU_FUNC_CACHE_PREFER_EQUAL
        else:
            config = driver.CUfunc_cache.CU_FUNC_CACHE_PREFER_NONE

        result = driver.cuFuncSetCacheConfig(self._handle, config)
        if result[0].value != 0:
            raise RuntimeError(f"cuFuncSetCacheConfig failed with error {result[0]}")

    def set_shared_memory_carveout(self, value):
        """
        Set the shared memory carveout percentage for this function.

        Args:
            value: Percentage of shared memory to carve out (0-100), or -1 for default

        Raises:
            ValueError: If value is not in valid range [-1, 100]
        """
        from cuda.bindings import driver

        if not isinstance(value, int) or value < -1 or value > 100:
            raise ValueError(f"Carveout value must be an integer in range [-1, 100], got {value}")

        attr = driver.CUfunction_attribute.CU_FUNC_ATTRIBUTE_PREFERRED_SHARED_MEMORY_CARVEOUT
        result = driver.cuFuncSetAttribute(self._handle, attr, value)
        if result[0].value != 0:
            raise RuntimeError(f"cuFuncSetAttribute failed with error {result[0]}")


class CodeLibrary:
    """
    Code library that holds compiled CUDA code and provides access to CUDA functions.

    This provides the same interface as numba-cuda's CUDACodeLibrary for accessing
    the compiled CUDA function with its configuration methods.
    """

    def __init__(self, cubin, func_name):
        """
        Initialize with compiled cubin and function name.

        Args:
            cubin: The compiled cubin bytes
            func_name: The mangled function name in the cubin
        """
        self._cubin = cubin
        self._func_name = func_name
        self._module = None
        self._cufunc_cache = {}

    def get_cufunc(self):
        """
        Get the CUDA function handle, loading the module if necessary.

        Returns:
            CUFunc: A wrapper around the CUfunction handle
        """
        from cuda.bindings import driver
        from numba_cuda_mlir.numba_cuda.cudadrv import devices

        # Get current device
        ctx = devices.get_context()
        device = ctx.device
        device_id = device.id

        # Check cache first
        if device_id in self._cufunc_cache:
            return self._cufunc_cache[device_id]

        # Load module from cubin
        result = driver.cuModuleLoadData(self._cubin)
        if result[0].value != 0:
            raise RuntimeError(f"cuModuleLoadData failed with error {result[0]}")
        self._module = result[1]

        # Get function from module
        result = driver.cuModuleGetFunction(self._module, self._func_name.encode())
        if result[0].value != 0:
            raise RuntimeError(
                f"cuModuleGetFunction failed for '{self._func_name}' with error {result[0]}"
            )
        handle = result[1]

        # Wrap and cache
        cufunc = CUFunc(handle)
        self._cufunc_cache[device_id] = cufunc

        return cufunc

    def get_kernel_attributes(self):
        """Query kernel resource usage attributes from the CUDA driver."""
        from cuda.bindings import driver

        cufunc = self.get_cufunc()
        handle = cufunc._handle

        def _get_attr(attr):
            err, val = driver.cuFuncGetAttribute(attr, handle)
            if err.value != 0:
                raise RuntimeError(f"cuFuncGetAttribute failed for {attr}: {err}")
            return int(val)

        return {
            "regs_per_thread": _get_attr(driver.CUfunction_attribute.CU_FUNC_ATTRIBUTE_NUM_REGS),
            "shared_mem_per_block": _get_attr(
                driver.CUfunction_attribute.CU_FUNC_ATTRIBUTE_SHARED_SIZE_BYTES
            ),
            "const_mem_size": _get_attr(
                driver.CUfunction_attribute.CU_FUNC_ATTRIBUTE_CONST_SIZE_BYTES
            ),
            "local_mem_per_thread": _get_attr(
                driver.CUfunction_attribute.CU_FUNC_ATTRIBUTE_LOCAL_SIZE_BYTES
            ),
            "max_threads_per_block": _get_attr(
                driver.CUfunction_attribute.CU_FUNC_ATTRIBUTE_MAX_THREADS_PER_BLOCK
            ),
        }


class CompileResult:
    def __init__(self, cres):
        self.cres = cres
        self._code_library = None

    def __getattr__(self, attr):
        if hasattr(self.cres, attr):
            return getattr(self.cres, attr)
        if attr == "ptx":
            ptx = self.cres.metadata.get("ptx", "")
            if ptx:
                return ptx
            if self.cres.metadata.get("ltoir"):
                from numba_cuda_mlir.mlir_optimization import get_ptx

                ptx = get_ptx(self.cres)
                self.cres.metadata["ptx"] = ptx
                return ptx
        if attr == "mlir_module":
            from numba_cuda_mlir.mlir_lowering import get_mlir_module_str

            return get_mlir_module_str(self.cres.metadata)
        if attr in self.cres.metadata.keys():
            return self.cres.metadata[attr]
        # Provide compatibility attributes
        if attr == "_codelibrary":
            # Lazily create the code library with cubin and func_name from metadata
            if self._code_library is None:
                cubin = self.cres.metadata.get("cubin")
                func_name = self.cres.metadata.get("func_name")
                if cubin is not None and func_name is not None:
                    self._code_library = CodeLibrary(cubin, func_name)
            return self._code_library
        if attr == "_type_annotation":
            return getattr(self.cres, "type_annotation", None)
        if attr == "cooperative":
            return getattr(
                self.cres,
                "use_cooperative",
                self.cres.metadata.get("use_cooperative", False),
            )
        if attr == "max_cooperative_grid_blocks":
            return self._max_cooperative_grid_blocks
        if attr == "inspect_asm":
            return self._inspect_asm
        if attr == "inspect_llvm":
            raise NotImplementedError(
                "inspect_llvm is not supported. Use inspect_asm() to inspect the PTX."
            )
        return getattr(self.cres, attr)

    def _ensure_kernel_attrs(self):
        """Lazily populate kernel resource attributes from the CUDA driver."""
        if "regs_per_thread" in self.cres.metadata:
            return
        lib = self._codelibrary
        if lib is not None:
            attrs = lib.get_kernel_attributes()
            self.cres.metadata.update(attrs)

    def _max_cooperative_grid_blocks(self, blockdim, dynsmemsize=0):
        import functools
        from numba_cuda_mlir.numba_cuda.cudadrv import devices

        ctx = devices.get_context()
        cufunc = self._codelibrary.get_cufunc()
        if isinstance(blockdim, tuple):
            blockdim = functools.reduce(lambda x, y: x * y, blockdim)
        active_per_sm = ctx.get_active_blocks_per_multiprocessor(cufunc, blockdim, dynsmemsize)
        sm_count = ctx.device.MULTIPROCESSOR_COUNT
        return active_per_sm * sm_count

    def _inspect_asm(self, cc=None):
        """Return PTX assembly from the compiled cubin metadata."""
        ptx = self.cres.metadata.get("ptx", "")
        if ptx:
            return ptx
        return self.cres.metadata.get("asm", "")


def _compile_and_optimize(pyfunc, sig=None, targetoptions=None):
    return _compile(pyfunc, sig, targetoptions, optimized=True)


def _compile_only(pyfunc, sig=None, targetoptions=None):
    """Compile to MLIR without running the optimization pipeline."""
    from numba_cuda_mlir.cuda import jit
    from numba_cuda_mlir import mlir_compiler
    from numba_cuda_mlir._whole_function_planners import _RequireLaunchConfig
    from numba_cuda_mlir.numba_cuda.core import sigutils

    dispatcher = pyfunc
    if not isinstance(dispatcher, MLIRDispatcher):
        kws = targetoptions or {}
        dispatcher = jit(dispatcher, **kws)
    else:
        pyfunc = dispatcher.py_func

    if sig is None:
        sig = to_numba_type(inspect.signature(pyfunc))

    argtypes, return_type = sigutils.normalize_signature(sig)
    try:
        cres = mlir_compiler.compile_mlir(
            dispatcher.py_func,
            return_type,
            argtypes,
            targetoptions=dispatcher.targetoptions,
        )
    except _RequireLaunchConfig:
        raise RuntimeError(
            "whole-function planner requires launch metadata; compile through a "
            "configured kernel launch"
        ) from None
    return CompileResult(cres)


def _compile(pyfunc, sig=None, targetoptions=None, optimized=True):
    from numba_cuda_mlir.cuda import jit

    dispatcher = pyfunc
    if not isinstance(dispatcher, MLIRDispatcher):
        kws = (targetoptions or {}).copy()
        output = kws.pop("output", kws.pop("_compile_output", None))
        if output is not None:
            kws["lto"] = output == "ltoir"
        # Don't call verify_target_options here - jit() will do it
        dispatcher = jit(dispatcher, **kws)
    else:
        pyfunc = dispatcher.py_func
        output = (
            targetoptions.get("output", targetoptions.get("_compile_output", None))
            if targetoptions is not None
            else None
        )
        if targetoptions is not None:
            dispatcher.targetoptions.update(
                {k: v for k, v in targetoptions.items() if k not in ("output", "_compile_output")}
            )

    if sig is None:
        sig = to_numba_type(inspect.signature(pyfunc))

    abi_info = targetoptions.get("abi_info", None) if targetoptions is not None else None
    cres = dispatcher.compile(sig, abi_info=abi_info, output=output)
    return CompileResult(cres)


def compile_for(func, *args):
    from numba_cuda_mlir.numba_cuda.typing.typeof import typeof

    sig = typing.signature(types.none, *[typeof(arg) for arg in args])
    cres = _compile_and_optimize(func, sig, {"lto": False})
    return cres


def compile_ptx(
    pyfunc,
    sig,
    debug=None,
    lineinfo=False,
    device=False,
    fastmath=False,
    cc=None,
    opt=None,
    abi="numba",
    abi_info=None,
    forceinline=False,
    launch_bounds=None,
):
    """Compile a Python function to PTX for a given signature."""
    return compile(
        pyfunc,
        sig,
        debug=debug,
        lineinfo=lineinfo,
        device=device,
        fastmath=fastmath,
        cc=cc,
        opt=opt,
        abi=abi,
        abi_info=abi_info,
        output="ptx",
        forceinline=forceinline,
        launch_bounds=launch_bounds,
    )


def compile(
    pyfunc,
    sig,
    debug=None,
    lineinfo=False,
    device=True,
    fastmath=False,
    cc=None,
    opt=None,
    abi="c",
    abi_info=None,
    output="ptx",
    forceinline=False,
    launch_bounds=None,
):
    """Compile a Python function to PTX or LTO-IR (numba-cuda compatible API)."""
    # Validate output type
    if output not in ("ptx", "ltoir"):
        raise NotImplementedError(f"Unsupported output type: {output}")

    # C ABI is not supported for kernels
    if abi == "c" and not device:
        raise NotImplementedError("The C ABI is not supported for kernels")

    # forceinline requires LTO-IR output
    if forceinline and output != "ltoir":
        raise ValueError("Can only designate forced inlining in LTO-IR")

    targetoptions = {}
    if debug is not None:
        targetoptions["debug"] = debug
    if lineinfo:
        targetoptions["lineinfo"] = lineinfo
    if device:
        targetoptions["device"] = device
    if fastmath:
        targetoptions["fastmath"] = fastmath
    if cc is not None:
        targetoptions["cc"] = cc
    if opt is not None:
        targetoptions["opt"] = opt
    if abi != "c":
        targetoptions["abi"] = abi
    if abi_info is not None:
        targetoptions["abi_info"] = abi_info
    if forceinline:
        targetoptions["inline"] = "always"
    if launch_bounds is not None:
        targetoptions["launch_bounds"] = launch_bounds
    targetoptions["_compile_output"] = output
    targetoptions["lto"] = output == "ltoir"

    optimized = _compile_and_optimize(pyfunc, sig, targetoptions)

    # Kernels must have void return type
    if not device and optimized.signature.return_type != types.none:
        raise TypeError("CUDA kernel must have void return type")

    if output == "ptx":
        return optimized.ptx, optimized.signature.return_type
    elif output == "ltoir":
        return optimized.metadata.get("ltoir"), optimized.signature.return_type
    else:
        return optimized.ptx, optimized.signature.return_type


def compile_all(*args, **kwargs):
    raise NotImplementedError("To be implemented")


def compile_result(pyfunc, sig=None):
    """Compile and return full CompileResult for internal use."""
    return _compile_and_optimize(pyfunc, sig)


def compile_mlir(pyfunc, sig, optimized=False, **targetoptions):
    if optimized:
        cres = _compile_and_optimize(pyfunc, sig, targetoptions)
        return cres.metadata["mlir_module_optimized"]
    else:
        from numba_cuda_mlir.descriptor import mlir_target
        from numba_cuda_mlir.mlir_lowering import get_mlir_module_str

        mlir_target.ensure_initialized()
        cres = _compile_only(pyfunc, sig, targetoptions)
        return get_mlir_module_str(cres.metadata)


def compile_cubin(pyfunc, sig, **targetoptions):
    cres = _compile_and_optimize(pyfunc, sig, targetoptions)
    return cres.metadata["cubin"]


@lru_cache(maxsize=None)
def _get_typing_key_for_callable(callable: T, sig: typing.Signature) -> T:
    from numba_cuda_mlir.descriptor import mlir_target

    typingctx = mlir_target.typing_context
    targetctx = mlir_target.target_context

    class mlir_function_template(ConcreteTemplate):
        key = callable
        cases = [sig]

    typingctx.insert_user_function(callable, mlir_function_template)
    return mlir_function_template.key


def declare_mlir_library(
    source: str | Path,
) -> Callable[[...], Any]:
    """
    Declare a library defined in MLIR to be callable from a numba-cuda-mlir Python kernel.
    """
    from numba_cuda_mlir.typing.externals import ExternMLIRLibrary
    from numba_cuda_mlir.lowering_utilities.discover_functions import discover_functions

    if isinstance(source, Path):
        with open(source, "r") as f:
            source = f.read()

    functions = discover_functions(source)

    lib = ExternMLIRLibrary(functions, source)
    return lib


class ExternFunction:
    """A descriptor that can be used to call the external function from within
    a Python kernel.

    Note: This class intentionally does not define __hash__ or __eq__ so that
    each instance is unique by identity. This matches numba-cuda's behavior
    and prevents "key already in dictionary" errors when the same external
    function is declared multiple times.
    """

    def __init__(self, name, sig, use_cooperative=False, link=tuple(), abi="numba"):
        self.name = name
        self.sig = sig
        self.use_cooperative = use_cooperative
        self.link = link
        self.abi = abi


@typeof_impl.register(ExternFunction)
def typeof_extern_function(val, c):
    class device_function_template(ConcreteTemplate):
        key = val
        cases = [val.sig]

        def get_impl_key(self, sig):
            return ExternFunction

    return types.Function(device_function_template)


@lower_builtin(ExternFunction, types.VarArg(types.Any))
def extern_function_dummy_lowering(context, builder, sig, args):
    return context.get_dummy_value()


extern_function_dummy_lowering.__module__ = "numba_cuda_mlir.numba_cuda.compiler"


def declare_device_function(name, restype, argtypes, link, use_cooperative, abi="numba"):
    from numba_cuda_mlir.descriptor import mlir_target

    mlir_target.ensure_initialized()
    typingctx = mlir_target.typing_context
    targetctx = mlir_target.target_context
    sig = typing.signature(restype, *argtypes)

    extfn = ExternFunction(name, sig, use_cooperative, link, abi=abi)

    # TODO(ajm): move numba_cuda_mlir to use numba.cuda's typing infrastructure, then use that instead
    class device_function_template(ConcreteTemplate):
        key = extfn
        cases = [sig]

    # Typing
    typingctx.insert_user_function(extfn, device_function_template)

    return device_function_template


def declare_device(name, sig, link=None, use_cooperative=False, abi="numba"):
    """
    Declare the signature of a foreign function. Returns a descriptor that can
    be used to call the function from a Python kernel.

    :param name: The name of the foreign function.
    :type name: str
    :param sig: The Numba signature of the function.
    :param link: External code to link when calling the function.
    :param use_cooperative: External code requires cooperative launch.
    :param abi: The ABI to use for the function. "numba" for Numba ABI, "c" for C ABI.
    """
    if abi not in ("numba", "c"):
        raise NotImplementedError(f"Unsupported ABI: {abi}")

    if link is None:
        link = tuple()
    else:
        if not isinstance(link, (list, tuple, set)):
            link = (link,)

    argtypes, restype = sigutils.normalize_signature(sig)
    if restype is None:
        msg = "Return type must be provided for device declarations"
        raise TypeError(msg)

    template = declare_device_function(name, restype, argtypes, link, use_cooperative, abi)

    return template.key



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

import collections
import warnings
from functools import cached_property

from .abstract import DTypeSpec, IteratorType, MutableSequence, Number, Type
from .common import Buffer, Opaque, SimpleIteratorType
from numba_cuda_mlir.numba_cuda.typeconv import Conversion
from .misc import UnicodeType
from .containers import Bytes
import numpy as np


class CharSeq(Type):
    """
    A fixed-length 8-bit character sequence.
    """

    mutable = True

    def __init__(self, count):
        self.count = count
        name = "[char x %d]" % count
        super().__init__(name)

    @property
    def key(self):
        return self.count

    def can_convert_from(self, typingctx, other):
        if isinstance(other, Bytes):
            return Conversion.safe


class UnicodeCharSeq(Type):
    """
    A fixed-length unicode character sequence.
    """

    mutable = True

    def __init__(self, count):
        self.count = count
        name = "[unichr x %d]" % count
        super().__init__(name)

    @property
    def key(self):
        return self.count

    def can_convert_to(self, typingctx, other):
        if isinstance(other, UnicodeCharSeq):
            return Conversion.safe

    def can_convert_from(self, typingctx, other):
        if isinstance(other, UnicodeType):
            # Assuming that unicode_type itemsize is not greater than
            # numpy.dtype('U1').itemsize that UnicodeCharSeq is based
            # on.
            return Conversion.safe

    def __repr__(self):
        return f"UnicodeCharSeq({self.count})"


_RecordField = collections.namedtuple(
    "_RecordField",
    "type,offset,alignment,title",
)


class Record(Type):
    """
    A Record datatype can be mapped to a NumPy structured dtype.
    A record is very flexible since it is laid out as a list of bytes.
    Fields can be mapped to arbitrary points inside it, even if they overlap.

    *fields* is a list of `(name:str, data:dict)`.
        Where `data` is `{ type: Type, offset: int }`
    *size* is an int; the record size
    *aligned* is a boolean; whether the record is ABI aligned.
    """

    mutable = True

    def __init__(self, fields, size, aligned):
        fields = self._normalize_fields(fields)
        self.fields = dict(fields)
        self.size = size
        self.aligned = aligned

        # Create description
        descbuf = []
        fmt = "{}[type={};offset={}{}]"
        for k, infos in fields:
            extra = ""
            if infos.alignment is not None:
                extra += ";alignment={}".format(infos.alignment)
            elif infos.title is not None:
                extra += ";title={}".format(infos.title)
            descbuf.append(fmt.format(k, infos.type, infos.offset, extra))

        desc = ",".join(descbuf)
        name = "Record({};{};{})".format(desc, self.size, self.aligned)
        super().__init__(name)

        self.bitwidth = self.dtype.itemsize * 8

    @classmethod
    def _normalize_fields(cls, fields):
        """
        fields:
            [name: str,
             value: {
                 type: Type,
                 offset: int,
                 [ alignment: int ],
                 [ title : str],
             }]
        """
        res = []
        for name, infos in sorted(fields, key=lambda x: (x[1]["offset"], x[0])):
            fd = _RecordField(
                type=infos["type"],
                offset=infos["offset"],
                alignment=infos.get("alignment"),
                title=infos.get("title"),
            )
            res.append((name, fd))
        return res

    @property
    def key(self):
        # Numpy dtype equality doesn't always succeed, use the name instead
        # (https://github.com/numpy/numpy/issues/5715)
        return self.name

    @property
    def mangling_args(self):
        return self.__class__.__name__, (self._code,)

    def __len__(self):
        """Returns the number of fields"""
        return len(self.fields)

    def offset(self, key):
        """Get the byte offset of a field from the start of the structure."""
        return self.fields[key].offset

    def typeof(self, key):
        """Get the type of a field."""
        return self.fields[key].type

    def alignof(self, key):
        """Get the specified alignment of the field.

        Since field alignment is optional, this may return None.
        """
        return self.fields[key].alignment

    def has_titles(self):
        """Returns True the record uses titles."""
        return any(fd.title is not None for fd in self.fields.values())

    def is_title(self, key):
        """Returns True if the field named *key* is a title."""
        return self.fields[key].title == key

    @property
    def members(self):
        """An ordered list of (name, type) for the fields."""
        ordered = sorted(self.fields.items(), key=lambda x: x[1].offset)
        return [(k, v.type) for k, v in ordered]

    @property
    def dtype(self):
        from numba_cuda_mlir.numba_cuda.np.numpy_support import as_struct_dtype

        return as_struct_dtype(self)

    def can_convert_to(self, typingctx, other):
        """
        Convert this Record to the *other*.

        This method only implements width subtyping for records.
        """
        from numba_cuda_mlir.numba_cuda.core.errors import (
            NumbaExperimentalFeatureWarning,
        )

        if isinstance(other, Record):
            if len(other.fields) > len(self.fields):
                return
            for other_fd, self_fd in zip(other.fields.items(), self.fields.items()):
                if not other_fd == self_fd:
                    return
            warnings.warn(
                f"{self} has been considered a subtype of {other} "
                f" This is an experimental feature.",
                category=NumbaExperimentalFeatureWarning,
            )
            return Conversion.safe

    def __repr__(self):
        fields = [
            f"('{f_name}', "
            + f"{{'type': {repr(f_info.type)}, "
            + f"'offset': {f_info.offset}, "
            + f"'alignment': {f_info.alignment}, "
            + f"'title': {f_info.title}, "
            + "}"
            + ")"
            for f_name, f_info in self.fields.items()
        ]
        fields = "[" + ", ".join(fields) + "]"
        return f"Record({fields}, {self.size}, {self.aligned})"


class DType(DTypeSpec, Opaque):
    """
    Type class associated with the `np.dtype`.

    i.e. :code:`assert type(np.dtype('int32')) == np.dtype`

    np.dtype('int32')
    """

    def __init__(self, dtype):
        assert isinstance(dtype, Type)
        self._dtype = dtype
        name = "dtype(%s)" % (dtype,)
        super().__init__(name)

    @property
    def key(self):
        return self.dtype

    @property
    def dtype(self):
        return self._dtype

    def __getitem__(self, arg):
        res = super().__getitem__(arg)
        return res.copy(dtype=self.dtype)


class NumpyFlatType(SimpleIteratorType, MutableSequence):
    """
    Type class for `ndarray.flat()` objects.
    """

    def __init__(self, arrty):
        self.array_type = arrty
        yield_type = arrty.dtype
        self.dtype = yield_type
        name = "array.flat({arrayty})".format(arrayty=arrty)
        super().__init__(name, yield_type)

    @property
    def key(self):
        return self.array_type


class NumpyNdEnumerateType(SimpleIteratorType):
    """
    Type class for `np.ndenumerate()` objects.
    """

    def __init__(self, arrty):
        from . import Tuple, UniTuple, intp

        self.array_type = arrty
        yield_type = Tuple((UniTuple(intp, arrty.ndim), arrty.dtype))
        name = "ndenumerate({arrayty})".format(arrayty=arrty)
        super().__init__(name, yield_type)

    @property
    def key(self):
        return self.array_type


class NumpyNdIterType(IteratorType):
    """
    Type class for `np.nditer()` objects.

    The layout denotes in which order the logical shape is iterated on.
    "C" means logical order (corresponding to in-memory order in C arrays),
    "F" means reverse logical order (corresponding to in-memory order in
    F arrays).
    """

    def __init__(self, arrays):
        # Note inputs arrays can also be scalars, in which case they are
        # broadcast.
        self.arrays = tuple(arrays)
        self.layout = self._compute_layout(self.arrays)
        self.dtypes = tuple(getattr(a, "dtype", a) for a in self.arrays)
        self.ndim = max(getattr(a, "ndim", 0) for a in self.arrays)
        name = "nditer(ndim={ndim}, layout={layout}, inputs={arrays})".format(
            ndim=self.ndim, layout=self.layout, arrays=self.arrays
        )
        super().__init__(name)

    @classmethod
    def _compute_layout(cls, arrays):
        c = collections.Counter()
        for a in arrays:
            if not isinstance(a, Array):
                continue
            if a.layout in "CF" and a.ndim == 1:
                c["C"] += 1
                c["F"] += 1
            elif a.ndim >= 1:
                c[a.layout] += 1
        return "F" if c["F"] > c["C"] else "C"

    @property
    def key(self):
        return self.arrays

    @property
    def views(self):
        """
        The views yielded by the iterator.
        """
        return [Array(dtype, 0, "C") for dtype in self.dtypes]

    @property
    def yield_type(self):
        from . import BaseTuple

        views = self.views
        if len(views) > 1:
            return BaseTuple.from_types(views)
        else:
            return views[0]

    @cached_property
    def indexers(self):
        """
        A list of (kind, start_dim, end_dim, indices) where:
        - `kind` is either "flat", "indexed", "0d" or "scalar"
        - `start_dim` and `end_dim` are the dimension numbers at which
          this indexing takes place
        - `indices` is the indices of the indexed arrays in self.arrays
        """
        d = collections.OrderedDict()
        layout = self.layout
        ndim = self.ndim
        assert layout in "CF"
        for i, a in enumerate(self.arrays):
            if not isinstance(a, Array):
                indexer = ("scalar", 0, 0)
            elif a.ndim == 0:
                indexer = ("0d", 0, 0)
            else:
                if a.layout == layout or (a.ndim == 1 and a.layout in "CF"):
                    kind = "flat"
                else:
                    kind = "indexed"
                if layout == "C":
                    # If iterating in C order, broadcasting is done on the outer indices
                    indexer = (kind, ndim - a.ndim, ndim)
                else:
                    indexer = (kind, 0, a.ndim)
            d.setdefault(indexer, []).append(i)
        return list(k + (v,) for k, v in d.items())

    @cached_property
    def need_shaped_indexing(self):
        """
        Whether iterating on this iterator requires keeping track of
        individual indices inside the shape.  If False, only a single index
        over the equivalent flat shape is required, which can make the
        iterator more efficient.
        """
        for kind, start_dim, end_dim, _ in self.indexers:
            if kind in ("0d", "scalar"):
                pass
            elif kind == "flat":
                if (start_dim, end_dim) != (0, self.ndim):
                    # Broadcast flat iteration needs shaped indexing
                    # to know when to restart iteration.
                    return True
            else:
                return True
        return False


class NumpyNdIndexType(SimpleIteratorType):
    """
    Type class for `np.ndindex()` objects.
    """

    def __init__(self, ndim):
        from . import UniTuple, intp

        self.ndim = ndim
        yield_type = UniTuple(intp, self.ndim)
        name = "ndindex(ndim={ndim})".format(ndim=ndim)
        super().__init__(name, yield_type)

    @property
    def key(self):
        return self.ndim


class Array(Buffer):
    """
    Type class for Numpy arrays.
    """

    def __init__(self, dtype, ndim, layout, readonly=False, name=None, aligned=True):
        if readonly:
            self.mutable = False
        if not aligned or (isinstance(dtype, Record) and not dtype.aligned):
            self.aligned = False
        if isinstance(dtype, NestedArray):
            ndim += dtype.ndim
            dtype = dtype.dtype
        if name is None:
            type_name = "array"
            if not self.mutable:
                type_name = "readonly " + type_name
            if not self.aligned:
                type_name = "unaligned " + type_name
            name = "%s(%s, %sd, %s)" % (type_name, dtype, ndim, layout)
        super().__init__(dtype, ndim, layout, name=name)

    @property
    def mangling_args(self):
        args = [
            self.dtype,
            self.ndim,
            self.layout,
            "mutable" if self.mutable else "readonly",
            "aligned" if self.aligned else "unaligned",
        ]
        return self.__class__.__name__, args

    def copy(self, dtype=None, ndim=None, layout=None, readonly=None):
        if dtype is None:
            dtype = self.dtype
        if ndim is None:
            ndim = self.ndim
        if layout is None:
            layout = self.layout
        if readonly is None:
            readonly = not self.mutable
        return Array(
            dtype=dtype,
            ndim=ndim,
            layout=layout,
            readonly=readonly,
            aligned=self.aligned,
        )

    @property
    def key(self):
        return self.dtype, self.ndim, self.layout, self.mutable, self.aligned

    def unify(self, typingctx, other):
        """
        Unify this with the *other* Array.
        """
        # If other is array and the ndim matches
        if isinstance(other, Array) and other.ndim == self.ndim:
            # If dtype matches or other.dtype is undefined (inferred)
            if other.dtype == self.dtype or not other.dtype.is_precise():
                if self.layout == other.layout:
                    layout = self.layout
                else:
                    layout = "A"
                readonly = not (self.mutable and other.mutable)
                aligned = self.aligned and other.aligned
                return Array(
                    dtype=self.dtype,
                    ndim=self.ndim,
                    layout=layout,
                    readonly=readonly,
                    aligned=aligned,
                )

    def can_convert_to(self, typingctx, other):
        """
        Convert this Array to the *other*.
        """
        if isinstance(other, Array) and other.ndim == self.ndim and other.dtype == self.dtype:
            if (
                other.layout in ("A", self.layout)
                and (self.mutable or not other.mutable)
                and (self.aligned or not other.aligned)
            ):
                return Conversion.safe

    def is_precise(self):
        return self.dtype.is_precise()

    @property
    def box_type(self):
        """Returns the Python type to box to."""
        return np.ndarray

    def __repr__(self):
        return (
            f"Array({repr(self.dtype)}, {self.ndim}, '{self.layout}', "
            f"{not self.mutable}, aligned={self.aligned})"
        )


class ArrayCTypes(Type):
    """
    This is the type for `np.ndarray.ctypes`.
    """

    def __init__(self, arytype):
        # This depends on the ndim for the shape and strides attributes,
        # even though they are not implemented, yet.
        self.dtype = arytype.dtype
        self.ndim = arytype.ndim
        name = "ArrayCTypes(dtype={0}, ndim={1})".format(self.dtype, self.ndim)
        super().__init__(name)

    @property
    def key(self):
        return self.dtype, self.ndim

    def can_convert_to(self, typingctx, other):
        """
        Convert this type to the corresponding pointer type.
        This allows passing a array.ctypes object to a C function taking
        a raw pointer.

        Note that in pure Python, the array.ctypes object can only be
        passed to a ctypes function accepting a c_void_p, not a typed
        pointer.
        """
        from . import CPointer, voidptr

        # XXX what about readonly
        if isinstance(other, CPointer) and other.dtype == self.dtype:
            return Conversion.safe
        elif other == voidptr:
            return Conversion.safe


class ArrayFlags(Type):
    """
    This is the type for `np.ndarray.flags`.
    """

    def __init__(self, arytype):
        self.array_type = arytype
        name = "ArrayFlags({0})".format(self.array_type)
        super().__init__(name)

    @property
    def key(self):
        return self.array_type


class NestedArray(Array):
    """
    A NestedArray is an array nested within a structured type (which are "void"
    type in NumPy parlance). Unlike an Array, the shape, and not just the number
    of dimensions is part of the type of a NestedArray.
    """

    def __init__(self, dtype, shape):
        if isinstance(dtype, NestedArray):
            tmp = Array(dtype.dtype, dtype.ndim, "C")
            shape += dtype.shape
            dtype = tmp.dtype
        assert dtype.bitwidth % 8 == 0, "Dtype bitwidth must be a multiple of bytes"
        self._shape = shape
        name = "nestedarray(%s, %s)" % (dtype, shape)
        ndim = len(shape)
        super().__init__(dtype, ndim, "C", name=name)

    @property
    def shape(self):
        return self._shape

    @property
    def nitems(self):
        l = 1
        for s in self.shape:
            l = l * s
        return l

    @property
    def size(self):
        return self.dtype.bitwidth // 8

    @property
    def strides(self):
        stride = self.size
        strides = []
        for i in reversed(self._shape):
            strides.append(stride)
            stride *= i
        return tuple(reversed(strides))

    @property
    def key(self):
        return self.dtype, self.shape

    def __repr__(self):
        return f"NestedArray({repr(self.dtype)}, {self.shape})"


class NumPyRandomBitGeneratorType(Type):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.name = "NumPyRandomBitGeneratorType"


class NumPyRandomGeneratorType(Type):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.name = "NumPyRandomGeneratorType"


class PolynomialType(Type):
    def __init__(self, coef, domain=None, window=None, n_args=1):
        super().__init__(name=f"PolynomialType({coef}, {domain}, {domain}, {n_args})")
        self.coef = coef
        self.domain = domain
        self.window = window
        # We use n_args to keep track of the number of arguments in the
        # constructor, since the types of domain and window arguments depend on
        # that and we need that information when boxing
        self.n_args = n_args



# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
import operator
from numba_cuda_mlir.lowering_utilities import index_of
from numba_cuda_mlir.numba_cuda import types
from numba_cuda_mlir.numba_cuda.types.ext_types import Dim3
from numba_cuda_mlir.mlir_lowering import MLIRLower
from numba_cuda_mlir._mlir import ir
from numba_cuda_mlir._mlir.extras import types as T
from numba_cuda_mlir._mlir.dialects import gpu, memref, arith
from numba_cuda_mlir.mlir.dialect_exts import scf
from numba_cuda_mlir.mlir.dialect_exts.scf import (
    if_ctx_manager as if_,
    else_ctx_manager as else_,
)
from numba_cuda_mlir.mlir.dialect_exts import llvm
from numba_cuda_mlir.lowering_registry import LoweringRegistry

registry = LoweringRegistry()
lower = registry.lower
from numba_cuda_mlir.lowering_utilities import (
    get_or_insert_function,
    lookup_callee_in_module,
    tensor_to_memref,
    memref_to_tensor,
    index_of,
    int_of,
)
from functools import singledispatch


def _dtype_format_string(dtype: ir.Type | type) -> str:
    match dtype:
        case ir.IntegerType():
            return "%d"
        case ir.FloatType():
            return "%f"
        case _ if dtype in (int, bool):
            return "%d"
        case _ if dtype is float:
            return "%f"
        case _ if dtype is str:
            return "%s"
        case _:
            raise ValueError(f"Unsupported type: {dtype}")


@singledispatch
def _lower_literal_print(value: int | float | bool | str) -> None:
    raise NotImplementedError(f"Unsupported type: {type(value)}")


@_lower_literal_print.register(int)
def _(value) -> None:
    mlir_value = arith.constant(T.i64(), value)
    return gpu.printf("%ld", mlir_value)


@_lower_literal_print.register(bool)
def _(value) -> None:
    return gpu.printf("True" if value else "False")


@_lower_literal_print.register(float)
def _(value) -> None:
    mlir_value = arith.constant(T.f64(), value)
    return gpu.printf("%f", mlir_value)


@_lower_literal_print.register(str)
def _(value) -> None:
    return gpu.printf(value)


def _memref_runtime_type_name(dtype: ir.Type) -> str:
    match dtype:
        case ir.IntegerType():
            return "I" + str(dtype.width)
        case ir.BF16Type():
            return "BF16"
        case ir.FloatType():
            return "F" + str(dtype.width)
        case _:
            raise ValueError(f"Unsupported type: {dtype}")


def _lower_memref_print(mlir_lower: MLIRLower, value: ir.Value) -> None:
    dtype = value.type.element_type
    value = tensor_to_memref(value)
    vt: ir.MemRefType = value.type
    if not vt.has_rank:
        raise TypeError("Cannot print unranked memref")
    rank = vt.rank
    c0 = index_of(0)
    c1 = index_of(1)
    gpu.printf(f"memref<shape=[")
    dims = []
    for iv in range(rank):
        dim = memref.dim(value, index_of(iv))
        dims.append(dim)
        gpu.printf("%d,", dim)

    # printing the data is probably not a good idea anyways (look at what pytorch tensor's
    # __str__ does) and it probably needs to be recursive and therefore defined in the
    # runtime...

    # gpu.printf("],data=[")
    # linear_mr = memref_to_tensor(value)
    # if rank > 1:
    #     mr_len = list(itertools.accumulate(dims, initial=1, func=operator.mul))
    #     linear_mr_dim = mr_len[-1]
    #     linear_mr_type = T.tensor(ir.RankedTensorType.get_dynamic_size(), dtype)
    #     linear_mr = tensor.collapse_shape(
    #         linear_mr_type, linear_mr, [list(range(rank))]
    #     )
    # else:
    #     linear_mr_dim = tensor.dim(linear_mr, index_of(0))

    # @scf.for__(c0, linear_mr_dim, c1)
    # def for_op(iv: T.index()):
    #     value_at_index = tensor.extract(linear_mr, [iv])
    #     gpu.printf("%d,", value_at_index)

    gpu.printf("]>")


def _lower_variable_print(mlir_lower: MLIRLower, value: ir.Value) -> None:
    match value.type:
        case ir.IntegerType() as int_type if int_type.width == 1:
            # Boolean (i1) - conditionally print "True" or "False"
            with if_(value, results=[]) as if_op:
                gpu.printf("True")
                scf.yield_([])
            with else_(if_op):
                gpu.printf("False")
                scf.yield_([])
        case ir.IntegerType():
            return gpu.printf(_dtype_format_string(value.type), value)
        case ir.BF16Type():
            # Convert bf16 to f32 for printing
            f32_val = arith.extf(T.f32(), value)
            return gpu.printf("%f", f32_val)
        case ir.F16Type():
            # Convert f16 to f32 for printing
            f32_val = arith.extf(T.f32(), value)
            return gpu.printf("%f", f32_val)
        case ir.FloatType():
            return gpu.printf(_dtype_format_string(value.type), value)
        case ir.MemRefType() | ir.RankedTensorType():
            return _lower_memref_print(mlir_lower, value)
        case _ if str(value.type).startswith("!llvm.struct"):
            _lower_string_struct_print(value)
        case _:
            raise ValueError(f"Unsupported type: {value.type}")


def _lower_string_struct_print(value: ir.Value) -> None:
    """Print a unicode_type string struct atomically via ``%s``.

    The struct layout is ``(ptr, i64, i32, i32, i64, ptr, ptr)`` where
    field 0 is the data pointer (to a null-terminated byte array).

    gpu.printf only accepts integer/float operands, so we pass the pointer
    as an i64 and rely on CUDA's printf ``%s`` implementation to interpret
    the integer as a device pointer.
    """
    data_ptr = llvm.extractvalue(llvm.PointerType.get(), value, [0])
    ptr_as_i64 = llvm.ptrtoint(res=T.i64(), arg=data_ptr)
    gpu.printf("%s", ptr_as_i64)


def _lower_tuple_print(mlir_lower: MLIRLower, values: list, sep: str = " ") -> None:
    """Print a tuple in format (v1, v2, v3) or (v1,) for single-element."""
    gpu.printf("(")
    for i, val in enumerate(values):
        if i > 0:
            gpu.printf(", ")
        _lower_print_arg(mlir_lower, val)
    if len(values) == 1:
        gpu.printf(",")
    gpu.printf(")")


def _lower_dim3_print(dim3_obj) -> None:
    """Print a Dim3 in format (x, y, z)."""
    from numba_cuda_mlir import cuda

    # Determine which dim3 object this is
    if dim3_obj == cuda.threadIdx:
        x_val = gpu.thread_id(gpu.Dimension.x)
        y_val = gpu.thread_id(gpu.Dimension.y)
        z_val = gpu.thread_id(gpu.Dimension.z)
    elif dim3_obj == cuda.blockIdx:
        x_val = gpu.block_id(gpu.Dimension.x)
        y_val = gpu.block_id(gpu.Dimension.y)
        z_val = gpu.block_id(gpu.Dimension.z)
    elif dim3_obj == cuda.blockDim:
        x_val = gpu.block_dim(gpu.Dimension.x)
        y_val = gpu.block_dim(gpu.Dimension.y)
        z_val = gpu.block_dim(gpu.Dimension.z)
    elif dim3_obj == cuda.gridDim:
        x_val = gpu.grid_dim(gpu.Dimension.x)
        y_val = gpu.grid_dim(gpu.Dimension.y)
        z_val = gpu.grid_dim(gpu.Dimension.z)
    else:
        raise ValueError(f"Unknown Dim3 object: {dim3_obj}")

    gpu.printf("(")
    gpu.printf("%d", x_val)
    gpu.printf(", ")
    gpu.printf("%d", y_val)
    gpu.printf(", ")
    gpu.printf("%d", z_val)
    gpu.printf(")")


def _is_dim3_object(arg) -> bool:
    """Check if arg is a Dim3 object (threadIdx, blockIdx, etc.)."""
    from numba_cuda_mlir import cuda

    return arg in (
        cuda.threadIdx,
        cuda.blockIdx,
        cuda.blockDim,
        cuda.gridDim,
    )


def _lower_print_arg(mlir_lower: MLIRLower, arg) -> None:
    """Lower a single print argument."""
    if _is_dim3_object(arg):
        _lower_dim3_print(arg)
    elif isinstance(arg, bool):
        _lower_literal_print(arg)
    elif isinstance(arg, (int, float, str)):
        _lower_literal_print(arg)
    elif isinstance(arg, (list, tuple)):
        _lower_tuple_print(mlir_lower, arg)
    elif isinstance(arg, ir.Value):
        _lower_variable_print(mlir_lower, arg)
    else:
        raise ValueError(f"Unsupported print argument type: {type(arg)}")


def _get_format_and_values(mlir_lower: MLIRLower, arg) -> tuple[str, list[ir.Value]]:
    """Get format string piece and values for a single argument.

    Returns (format_string, [values]) where values are ir.Values to pass to printf.
    For complex types that can't be handled in a single printf, returns (None, [])
    to indicate fallback to separate printf calls is needed.
    """
    if _is_dim3_object(arg):
        from numba_cuda_mlir import cuda

        if arg == cuda.threadIdx:
            x = gpu.thread_id(gpu.Dimension.x)
            y = gpu.thread_id(gpu.Dimension.y)
            z = gpu.thread_id(gpu.Dimension.z)
        elif arg == cuda.blockIdx:
            x = gpu.block_id(gpu.Dimension.x)
            y = gpu.block_id(gpu.Dimension.y)
            z = gpu.block_id(gpu.Dimension.z)
        elif arg == cuda.blockDim:
            x = gpu.block_dim(gpu.Dimension.x)
            y = gpu.block_dim(gpu.Dimension.y)
            z = gpu.block_dim(gpu.Dimension.z)
        elif arg == cuda.gridDim:
            x = gpu.grid_dim(gpu.Dimension.x)
            y = gpu.grid_dim(gpu.Dimension.y)
            z = gpu.grid_dim(gpu.Dimension.z)
        return "(%lld, %lld, %lld)", [x, y, z]
    elif isinstance(arg, bool):
        return ("True" if arg else "False"), []
    elif isinstance(arg, int):
        return "%lld", [arith.constant(T.i64(), arg)]
    elif isinstance(arg, float):
        return "%f", [arith.constant(T.f64(), arg)]
    elif isinstance(arg, str):
        return arg, []
    elif isinstance(arg, (list, tuple)):
        # Build tuple format: (v1, v2, v3) or (v1,)
        fmt_parts = []
        values = []
        for i, val in enumerate(arg):
            if i > 0:
                fmt_parts.append(", ")
            sub_fmt, sub_vals = _get_format_and_values(mlir_lower, val)
            if sub_fmt is None:
                return None, []  # Can't handle in single printf
            fmt_parts.append(sub_fmt)
            values.extend(sub_vals)
        if len(arg) == 1:
            return "(" + "".join(fmt_parts) + ",)", values
        return "(" + "".join(fmt_parts) + ")", values
    elif isinstance(arg, ir.Value):
        match arg.type:
            case ir.IntegerType() as int_type if int_type.width == 1:
                return None, []
            case ir.IntegerType():
                return "%lld", [arg]
            case ir.BF16Type():
                return "%f", [arith.extf(T.f32(), arg)]
            case ir.F16Type():
                return "%f", [arith.extf(T.f32(), arg)]
            case ir.FloatType():
                return "%f", [arg]
            case _ if str(arg.type).startswith("!llvm.struct"):
                data_ptr = llvm.extractvalue(llvm.PointerType.get(), arg, [0])
                return "%s", [llvm.ptrtoint(res=T.i64(), arg=data_ptr)]
            case ir.MemRefType() | ir.RankedTensorType():
                return None, []
            case _:
                return None, []
    return None, []


@lower(print, types.VarArg(types.Any))
def lower_print(mlir_lower: MLIRLower, target, args, kwargs: list[tuple[str, ir.Value]]):
    def _kwarg_str(kw_list, name, default):
        """Extract a string keyword arg from its literal type."""
        for k, v in kw_list:
            if k == name:
                ty = mlir_lower.get_numba_type(v.name)
                if isinstance(ty, types.StringLiteral):
                    return ty.literal_value
                loaded = mlir_lower.load_var(v)
                return loaded if isinstance(loaded, str) else default
        return default

    end = _kwarg_str(kwargs, "end", "\n")
    sep = _kwarg_str(kwargs, "sep", " ")
    remaining_kwargs = {k: mlir_lower.load_var(v) for k, v in kwargs if k not in ("end", "sep")}
    args = [mlir_lower.load_var(arg) for arg in args]

    if remaining_kwargs:
        raise ValueError(f"Unsupported keyword arguments: {remaining_kwargs}")

    # Try to build a single format string for all arguments
    fmt_parts = []
    all_values = []
    can_combine = True

    for i, arg in enumerate(args):
        if i > 0:
            fmt_parts.append(sep)
        fmt, values = _get_format_and_values(mlir_lower, arg)
        if fmt is None:
            can_combine = False
            break
        fmt_parts.append(fmt)
        all_values.extend(values)

    if can_combine and args:
        # Emit single printf with combined format string
        format_str = "".join(fmt_parts) + end
        gpu.printf(format_str, *all_values)
    elif not args:
        # Empty print - just print newline
        gpu.printf(end)
    else:
        # Fall back to separate printf calls for complex cases
        for i, arg in enumerate(args):
            if i > 0:
                gpu.printf(sep)
            _lower_print_arg(mlir_lower, arg)
        gpu.printf(end)




# Autogenerated by mlir-tblgen; don't manually edit.

from ._ods_common import _cext as _ods_cext
from ._ods_common import (
    equally_sized_accessor as _ods_equally_sized_accessor,
    get_default_loc_context as _ods_get_default_loc_context,
    get_op_results_or_values as _get_op_results_or_values,
    segmented_accessor as _ods_segmented_accessor,
)
_ods_ir = _ods_cext.ir
_ods_cext.globals.register_traceback_file_exclusion(__file__)

import builtins
from typing import Any as _Any, Sequence as _Sequence, Union as _Union, Optional as _Optional
import sys as _sys
if _sys.version_info >= (3, 12):
  from collections.abc import Buffer as _Buffer  # pytype: disable=not-supported-yet
else:
  try:
    from typing_extensions import Buffer as _Buffer
  except ImportError:
    _Buffer = _Any


from ._transform_ops_gen import _Dialect

@_ods_cext.register_operation(_Dialect)
class AlternativesOp(_ods_ir.OpView):
  r"""
  This op represents a choice over which of its regions is to be used.
  
  When `selected_region` is provided, the semantics are that this op is to be
  substituted for by the selected region, meaning the region's results become
  the results of this op. Without a provided `selected_region`, the semantics
  are that this non-deterministic choice is yet to be resolved -- which in
  terms of the op's interpreted semantics is a failure.
  
  The `selected_region` argument is either an `IntegerAttr` or a param holding
  an `IntegerAttr`, which should provide a valid zero-based index with respect
  to the number of alternatives, i.e. regions.
  """

  OPERATION_NAME = "transform.tune.alternatives"

  _ODS_REGIONS = (0, False)

  def __init__(self, results_: _Sequence[_ods_ir.Type], name: _Union[_Any, _ods_ir.StringAttr], num_alternatives: int, *, selected_region_attr: _Optional[_Union[_Any, _ods_ir.IntegerAttr]] = None, selected_region_param: _Optional[_ods_ir.Value] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    if selected_region_param is not None: operands.append(selected_region_param)
    _ods_context = _ods_get_default_loc_context(loc)
    attributes["name"] = (name if (
    isinstance(name, _ods_ir.Attribute) or
    not _ods_ir.AttrBuilder.contains('Builtin_StringAttr')) else
      _ods_ir.AttrBuilder.get('Builtin_StringAttr')(name, context=_ods_context))
    if selected_region_attr is not None: attributes["selected_region_attr"] = (selected_region_attr if (
        isinstance(selected_region_attr, _ods_ir.Attribute) or
        not _ods_ir.AttrBuilder.contains('APIntAttr')) else
          _ods_ir.AttrBuilder.get('APIntAttr')(selected_region_attr, context=_ods_context))
    results = []
    results.extend(results_)
    _ods_successors = None
    regions = 0 + num_alternatives
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def selected_region_param(self) -> _Optional[_ods_ir.Value]:
    return None if len(self.operation.operands) < 1 else self.operation.operands[0]

  @builtins.property
  def name(self) -> _ods_ir.StringAttr:
    return self.operation.attributes["name"]

  @name.setter
  def name(self, value: _ods_ir.StringAttr):
    if value is None:
      raise ValueError("'None' not allowed as value for mandatory attributes")
    self.operation.attributes["name"] = value

  @builtins.property
  def selected_region_attr(self) -> _Optional[_ods_ir.IntegerAttr]:
    if "selected_region_attr" not in self.operation.attributes:
      return None
    return self.operation.attributes["selected_region_attr"]

  @selected_region_attr.setter
  def selected_region_attr(self, value: _Optional[_ods_ir.IntegerAttr]):
    if value is not None:
      self.operation.attributes["selected_region_attr"] = value
    elif "selected_region_attr" in self.operation.attributes:
      del self.operation.attributes["selected_region_attr"]

  @selected_region_attr.deleter
  def selected_region_attr(self):
    del self.operation.attributes["selected_region_attr"]

  @builtins.property
  def results_(self) -> _ods_ir.OpResultList:
    _ods_variadic_group_length = len(self.operation.results) - 1 + 1
    return self.operation.results[0:0 + _ods_variadic_group_length]

  @builtins.property
  def alternatives(self) -> _ods_ir.RegionSequence:
    return self.regions[0:]

@_ods_cext.register_op_adaptor(AlternativesOp)
class AlternativesOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "transform.tune.alternatives"

  @builtins.property
  def selected_region_param(self) -> _Optional[_ods_ir.Value]:
    return None if len(self.operands) < 1 else self.operands[0]

  @builtins.property
  def name(self) -> _ods_ir.StringAttr:
    return self.attributes["name"]

  @builtins.property
  def selected_region_attr(self) -> _Optional[_ods_ir.IntegerAttr]:
    if "selected_region_attr" not in self.attributes:
      return None
    return self.attributes["selected_region_attr"]

def tune_alternatives(results_: _Sequence[_ods_ir.Type], name: _Union[_Any, _ods_ir.StringAttr], num_alternatives: int, *, selected_region_attr: _Optional[_Union[_Any, _ods_ir.IntegerAttr]] = None, selected_region_param: _Optional[_ods_ir.Value] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _Union[_ods_ir.OpResult, _ods_ir.OpResultList, AlternativesOp]:
  op = AlternativesOp(results_=results_, name=name, num_alternatives=num_alternatives, selected_region_attr=selected_region_attr, selected_region_param=selected_region_param, loc=loc, ip=ip); results = op.results
  return results if len(results) > 1 else (results[0] if len(results) == 1 else op)

@_ods_cext.register_operation(_Dialect)
class KnobOp(_ods_ir.OpView):
  r"""
  Provides a representation for "tunables" within schedules.
  
  Each op represents a single tunable, which has a `name` and a set
  of valid `options` described by an attribute. Without a specified
  `selected` option, this op represents a non-deterministic choice
  that has yet to be resolved -- as such, the interpreter runtime
  semantics is to raise a failure.
  
  The non-deterministic choice is resolved through providing a
  `selected` attribute. When provided, the interpreter runtime
  semantics are to return the `selected` attribute as a param through
  the op's result.
  
  -----
  
  In case the `options` attribute is an `ArrayAttr`, the verifier
  checks that the provided `selected` attribute occurs in `options`.
  """

  OPERATION_NAME = "transform.tune.knob"

  _ODS_REGIONS = (0, True)

  def __init__(self, result: _ods_ir.Type, name: _Union[_Any, _ods_ir.StringAttr], options: _Union[_Any, _ods_ir.Attribute], *, selected: _Optional[_Union[_Any, _ods_ir.Attribute]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    _ods_context = _ods_get_default_loc_context(loc)
    attributes["name"] = (name if (
    isinstance(name, _ods_ir.Attribute) or
    not _ods_ir.AttrBuilder.contains('Builtin_StringAttr')) else
      _ods_ir.AttrBuilder.get('Builtin_StringAttr')(name, context=_ods_context))
    attributes["options"] = (options if (
    isinstance(options, _ods_ir.Attribute) or
    not _ods_ir.AttrBuilder.contains('AnyAttr')) else
      _ods_ir.AttrBuilder.get('AnyAttr')(options, context=_ods_context))
    if selected is not None: attributes["selected"] = (selected if (
        isinstance(selected, _ods_ir.Attribute) or
        not _ods_ir.AttrBuilder.contains('AnyAttr')) else
          _ods_ir.AttrBuilder.get('AnyAttr')(selected, context=_ods_context))
    results = []
    results.append(result)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def name(self) -> _ods_ir.StringAttr:
    return self.operation.attributes["name"]

  @name.setter
  def name(self, value: _ods_ir.StringAttr):
    if value is None:
      raise ValueError("'None' not allowed as value for mandatory attributes")
    self.operation.attributes["name"] = value

  @builtins.property
  def options(self) -> _ods_ir.Attribute:
    return self.operation.attributes["options"]

  @options.setter
  def options(self, value: _ods_ir.Attribute):
    if value is None:
      raise ValueError("'None' not allowed as value for mandatory attributes")
    self.operation.attributes["options"] = value

  @builtins.property
  def selected(self) -> _Optional[_ods_ir.Attribute]:
    if "selected" not in self.operation.attributes:
      return None
    return self.operation.attributes["selected"]

  @selected.setter
  def selected(self, value: _Optional[_ods_ir.Attribute]):
    if value is not None:
      self.operation.attributes["selected"] = value
    elif "selected" in self.operation.attributes:
      del self.operation.attributes["selected"]

  @selected.deleter
  def selected(self):
    del self.operation.attributes["selected"]

  @builtins.property
  def result(self) -> _ods_ir.OpResult:
    return self.operation.results[0]

@_ods_cext.register_op_adaptor(KnobOp)
class KnobOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "transform.tune.knob"

  @builtins.property
  def name(self) -> _ods_ir.StringAttr:
    return self.attributes["name"]

  @builtins.property
  def options(self) -> _ods_ir.Attribute:
    return self.attributes["options"]

  @builtins.property
  def selected(self) -> _Optional[_ods_ir.Attribute]:
    if "selected" not in self.attributes:
      return None
    return self.attributes["selected"]

def tune_knob(result: _ods_ir.Type, name: _Union[_Any, _ods_ir.StringAttr], options: _Union[_Any, _ods_ir.Attribute], *, selected: _Optional[_Union[_Any, _ods_ir.Attribute]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _ods_ir.OpResult:
  return KnobOp(result=result, name=name, options=options, selected=selected, loc=loc, ip=ip).result




# Autogenerated by mlir-tblgen; don't manually edit.

from ._ods_common import _cext as _ods_cext
from ._ods_common import (
    equally_sized_accessor as _ods_equally_sized_accessor,
    get_default_loc_context as _ods_get_default_loc_context,
    get_op_results_or_values as _get_op_results_or_values,
    segmented_accessor as _ods_segmented_accessor,
)
_ods_ir = _ods_cext.ir
_ods_cext.globals.register_traceback_file_exclusion(__file__)

import builtins
from typing import Any as _Any, Sequence as _Sequence, Union as _Union, Optional as _Optional
import sys as _sys
if _sys.version_info >= (3, 12):
  from collections.abc import Buffer as _Buffer  # pytype: disable=not-supported-yet
else:
  try:
    from typing_extensions import Buffer as _Buffer
  except ImportError:
    _Buffer = _Any


from ._transform_ops_gen import _Dialect

@_ods_cext.register_operation(_Dialect)
class MatchSparseInOut(_ods_ir.OpView):
  r"""
  Checks if the payload op has any sparse inputs and/or outputs.
  """

  OPERATION_NAME = "transform.sparse_tensor.match.sparse_inout"

  _ODS_REGIONS = (0, True)

  def __init__(self, result: _ods_ir.Type, target: _ods_ir.Value, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(target)
    _ods_context = _ods_get_default_loc_context(loc)
    results = []
    results.append(result)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def target(self) -> _ods_ir.Value:
    return self.operation.operands[0]

  @builtins.property
  def result(self) -> _ods_ir.OpResult:
    return self.operation.results[0]

@_ods_cext.register_op_adaptor(MatchSparseInOut)
class MatchSparseInOutAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "transform.sparse_tensor.match.sparse_inout"

  @builtins.property
  def target(self) -> _ods_ir.Value:
    return self.operands[0]

def sparse_tensor_match_sparse_inout(result: _ods_ir.Type, target: _ods_ir.Value, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _ods_ir.OpResult:
  return MatchSparseInOut(result=result, target=target, loc=loc, ip=ip).result



#  Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
#  See https://llvm.org/LICENSE.txt for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

from functools import partial
from typing import Optional, List

from ..ir import (
    Attribute,
    BF16Type,
    ComplexType,
    F16Type,
    F32Type,
    F64Type,
    Float4E2M1FNType,
    Float6E2M3FNType,
    Float6E3M2FNType,
    Float8E3M4Type,
    Float8E4M3B11FNUZType,
    Float8E4M3FNType,
    Float8E4M3Type,
    Float8E5M2Type,
    Float8E8M0FNUType,
    FloatTF32Type,
    FunctionType,
    IndexType,
    IntegerType,
    MemRefType,
    NoneType,
    OpaqueType,
    RankedTensorType,
    StridedLayoutAttr,
    StringAttr,
    TupleType,
    Type,
    UnrankedMemRefType,
    UnrankedTensorType,
    VectorType,
)

index = lambda: IndexType.get()


def i(width):
    return IntegerType.get_signless(width)


def si(width):
    return IntegerType.get_signed(width)


def ui(width):
    return IntegerType.get_unsigned(width)


bool = lambda: i(1)
i8 = lambda: i(8)
i16 = lambda: i(16)
i32 = lambda: i(32)
i64 = lambda: i(64)

si8 = lambda: si(8)
si16 = lambda: si(16)
si32 = lambda: si(32)
si64 = lambda: si(64)

ui8 = lambda: ui(8)
ui16 = lambda: ui(16)
ui32 = lambda: ui(32)
ui64 = lambda: ui(64)

f16 = lambda: F16Type.get()
f32 = lambda: F32Type.get()
tf32 = lambda: FloatTF32Type.get()
f64 = lambda: F64Type.get()
bf16 = lambda: BF16Type.get()

f8E5M2 = lambda: Float8E5M2Type.get()
f8E4M3 = lambda: Float8E4M3Type.get()
f8E4M3FN = lambda: Float8E4M3FNType.get()
f8E4M3B11FNUZ = lambda: Float8E4M3B11FNUZType.get()
f8E3M4 = lambda: Float8E3M4Type.get()
f4E2M1FN = lambda: Float4E2M1FNType.get()
f6E2M3FN = lambda: Float6E2M3FNType.get()
f6E3M2FN = lambda: Float6E3M2FNType.get()
f8E8M0FNU = lambda: Float8E8M0FNUType.get()

none = lambda: NoneType.get()


def complex(type):
    return ComplexType.get(type)


def opaque(dialect_namespace, type_data):
    return OpaqueType.get(dialect_namespace, type_data)


def _shaped(*shape, element_type: Type = None, type_constructor=None):
    if type_constructor is None:
        raise ValueError("shaped is an abstract base class - cannot be constructed.")
    if (element_type is None and shape and not isinstance(shape[-1], Type)) or (
        shape and isinstance(shape[-1], Type) and element_type is not None
    ):
        raise ValueError(
            f"Either element_type must be provided explicitly XOR last arg to tensor type constructor must be the element type."
        )
    if element_type is not None:
        type = element_type
        sizes = shape
    else:
        type = shape[-1]
        sizes = shape[:-1]
    if sizes:
        return type_constructor(sizes, type)
    else:
        return type_constructor(type)


def vector(
    *shape,
    element_type: Type = None,
    scalable: Optional[List[bool]] = None,
    scalable_dims: Optional[List[int]] = None,
):
    return _shaped(
        *shape,
        element_type=element_type,
        type_constructor=partial(
            VectorType.get, scalable=scalable, scalable_dims=scalable_dims
        ),
    )


def tensor(*shape, element_type: Type = None, encoding: Optional[str] = None):
    if encoding is not None:
        encoding = StringAttr.get(encoding)
    if not shape or (len(shape) == 1 and isinstance(shape[-1], Type)):
        if encoding is not None:
            raise ValueError("UnrankedTensorType does not support encoding.")
        return _shaped(
            *shape, element_type=element_type, type_constructor=UnrankedTensorType.get
        )
    return _shaped(
        *shape,
        element_type=element_type,
        type_constructor=partial(RankedTensorType.get, encoding=encoding),
    )


def memref(
    *shape,
    element_type: Type = None,
    memory_space: Optional[int] = None,
    layout: Optional[StridedLayoutAttr] = None,
):
    if memory_space is not None:
        memory_space = Attribute.parse(str(memory_space))
    if not shape or (len(shape) == 1 and isinstance(shape[-1], Type)):
        return _shaped(
            *shape,
            element_type=element_type,
            type_constructor=partial(UnrankedMemRefType.get, memory_space=memory_space),
        )
    return _shaped(
        *shape,
        element_type=element_type,
        type_constructor=partial(
            MemRefType.get, memory_space=memory_space, layout=layout
        ),
    )


def tuple(*elements):
    return TupleType.get_tuple(elements)


def function(*, inputs, results):
    return FunctionType.get(inputs, results)
